import copy

import pytest
from icupy import U_ICU_VERSION_MAJOR_NUM

if U_ICU_VERSION_MAJOR_NUM < 50:
    pytest.skip("ICU4C<50", allow_module_level=True)

from icupy import (
    FieldPosition, FieldPositionIterator, Format, Formattable, ICUError,
    Locale, ParsePosition, TimeZone, TimeZoneFormat, TimeZoneNames,
    UErrorCode, UTimeZoneFormatGMTOffsetPatternType,
    UTimeZoneFormatParseOption, UTimeZoneFormatStyle, UTimeZoneFormatTimeType,
    U_MILLIS_PER_HOUR as HOUR, UnicodeString,
)


def test_api():
    assert issubclass(TimeZoneFormat, Format)

    # static TimeZoneFormat *icu::TimeZoneFormat::createInstance(
    #       const Locale &locale,
    #       UErrorCode &status
    # )
    fmt1 = TimeZoneFormat.create_instance(Locale("en"))
    fmt2 = TimeZoneFormat.create_instance("en")
    fmt3 = TimeZoneFormat.create_instance(Locale("ja"))
    assert isinstance(fmt1, TimeZoneFormat)
    assert isinstance(fmt2, TimeZoneFormat)
    assert isinstance(fmt3, TimeZoneFormat)

    # UBool icu::Format::operator!=(const Format &other)
    assert not (fmt1 != fmt2)
    assert fmt1 != fmt3
    assert fmt2 != fmt3

    # UBool icu::TimeZoneFormat::operator==(const Format &other)
    assert fmt1 == fmt2
    assert not (fmt1 == fmt3)
    assert not (fmt2 == fmt3)

    # UnicodeString &icu::TimeZoneFormat::formatOffsetLocalizedGMT(
    #       int32_t offset,
    #       UnicodeString &result,
    #       UErrorCode &status
    # )
    result = UnicodeString()
    string = fmt1.format_offset_localized_gmt(int(-8.5 * HOUR), result)
    assert isinstance(string, UnicodeString)
    assert id(result) == id(string)
    assert result == "GMT-08:30"

    # uint32_t icu::TimeZoneFormat::getDefaultParseOptions(void)
    assert (fmt2.get_default_parse_options()
            == UTimeZoneFormatParseOption.UTZFMT_PARSE_OPTION_NONE)

    # void icu::TimeZoneFormat::setDefaultParseOptions(uint32_t flags)
    fmt2.set_default_parse_options(
        UTimeZoneFormatParseOption.UTZFMT_PARSE_OPTION_ALL_STYLES)
    assert (fmt2.get_default_parse_options()
            == UTimeZoneFormatParseOption.UTZFMT_PARSE_OPTION_ALL_STYLES)

    # UnicodeString &icu::TimeZoneFormat::getGMTOffsetDigits(
    #       UnicodeString &digits
    # )
    digits = UnicodeString()
    result = fmt2.get_gmt_offset_digits(digits)
    assert isinstance(result, UnicodeString)
    assert id(result) == id(digits)
    assert result == "0123456789"

    # void icu::TimeZoneFormat::setGMTOffsetDigits(
    #       const UnicodeString &digits,
    #       UErrorCode &status
    # )
    fmt2.set_gmt_offset_digits(UnicodeString("abcdefghij"))
    assert fmt2.get_gmt_offset_digits(digits) == "abcdefghij"

    fmt2.set_gmt_offset_digits("0123456789")
    assert fmt2.get_gmt_offset_digits(digits) == "0123456789"

    # UnicodeString &icu::TimeZoneFormat::getGMTOffsetPattern(
    #       UTimeZoneFormatGMTOffsetPatternType type,
    #       UnicodeString &pattern
    # )
    pattern = UnicodeString()
    result = fmt2.get_gmt_offset_pattern(
        UTimeZoneFormatGMTOffsetPatternType.UTZFMT_PAT_POSITIVE_HM,
        pattern)
    assert isinstance(result, UnicodeString)
    assert id(result) == id(pattern)
    assert result == "+HH:mm"

    # void icu::TimeZoneFormat::setGMTOffsetPattern(
    #       UTimeZoneFormatGMTOffsetPatternType type,
    #       const UnicodeString &pattern,
    #       UErrorCode &status
    # )
    fmt2.set_gmt_offset_pattern(
        UTimeZoneFormatGMTOffsetPatternType.UTZFMT_PAT_POSITIVE_HM,
        UnicodeString("+H:mm"))
    assert fmt2.get_gmt_offset_pattern(
        UTimeZoneFormatGMTOffsetPatternType.UTZFMT_PAT_POSITIVE_HM,
        pattern) == "+H:mm"

    fmt2.set_gmt_offset_pattern(
        UTimeZoneFormatGMTOffsetPatternType.UTZFMT_PAT_POSITIVE_HM,
        "+HH:mm")
    assert fmt2.get_gmt_offset_pattern(
        UTimeZoneFormatGMTOffsetPatternType.UTZFMT_PAT_POSITIVE_HM,
        pattern) == "+HH:mm"

    # UnicodeString &icu::TimeZoneFormat::getGMTPattern(UnicodeString &pattern)
    result = fmt2.get_gmt_pattern(pattern)
    assert isinstance(result, UnicodeString)
    assert id(result) == id(pattern)
    assert result == "GMT{0}"

    # void icu::TimeZoneFormat::setGMTPattern(
    #       const UnicodeString &pattern,
    #       UErrorCode &status
    # )
    fmt2.set_gmt_pattern(UnicodeString("GMT {0}"))
    assert fmt2.get_gmt_pattern(pattern) == "GMT {0}"

    fmt2.set_gmt_pattern("GMT{0}")
    assert fmt2.get_gmt_pattern(pattern) == "GMT{0}"

    # UnicodeString &icu::TimeZoneFormat::getGMTZeroFormat(
    #       UnicodeString &gmtZeroFormat
    # )
    gmt_zero_format = UnicodeString()
    result = fmt2.get_gmt_zero_format(gmt_zero_format)
    assert isinstance(result, UnicodeString)
    assert id(result) == id(gmt_zero_format)
    assert result == "GMT"

    # void icu::TimeZoneFormat::setGMTZeroFormat(
    #       const UnicodeString &gmtZeroFormat,
    #       UErrorCode &status
    # )
    fmt2.set_gmt_zero_format(UnicodeString("UTC"))
    assert fmt2.get_gmt_zero_format(gmt_zero_format) == "UTC"

    fmt2.set_gmt_zero_format("GMT")
    assert fmt2.get_gmt_zero_format(gmt_zero_format) == "GMT"

    # const TimeZoneNames *icu::TimeZoneFormat::getTimeZoneNames()
    tzn = fmt2.get_time_zone_names()
    assert isinstance(tzn, TimeZoneNames)

    # void icu::TimeZoneFormat::setTimeZoneNames(const TimeZoneNames &tznames)
    tznames = TimeZoneNames.create_instance("ja")
    assert tznames != tzn
    fmt2.set_time_zone_names(tznames)
    assert fmt2.get_time_zone_names() == tznames

    # int32_t icu::TimeZoneFormat::parseOffsetISO8601(
    #       const UnicodeString &text,
    #       ParsePosition &pos
    # )
    pos = ParsePosition(0)
    result = fmt1.parse_offset_iso8601(UnicodeString("-08:30"), pos)
    assert pos.get_error_index() == -1
    assert result == -8.5 * HOUR

    pos = ParsePosition(0)
    result = fmt1.parse_offset_iso8601("-08:30", pos)
    assert pos.get_error_index() == -1
    assert result == -8.5 * HOUR

    # int32_t icu::TimeZoneFormat::parseOffsetLocalizedGMT(
    #       const UnicodeString &text,
    #       ParsePosition &pos
    # )
    pos = ParsePosition(0)
    result = fmt1.parse_offset_localized_gmt(UnicodeString("GMT-8:30"), pos)
    assert pos.get_error_index() == -1
    assert result == -8.5 * HOUR

    pos = ParsePosition(0)
    result = fmt1.parse_offset_localized_gmt("GMT-8:30", pos)
    assert pos.get_error_index() == -1
    assert result == -8.5 * HOUR


def test_clone():
    # TimeZoneFormat *icu::TimeZoneFormat::clone()
    fmt1 = TimeZoneFormat.create_instance("en")

    fmt2 = fmt1.clone()
    assert isinstance(fmt2, TimeZoneFormat)
    assert fmt2 == fmt1

    fmt3 = copy.copy(fmt1)
    assert fmt3 == fmt1

    fmt4 = copy.deepcopy(fmt1)
    assert fmt4 == fmt1


def test_format():
    fmt = TimeZoneFormat.create_instance("en")
    zone = TimeZone.create_time_zone("Asia/Tokyo")
    assert zone.get_raw_offset() == 9 * HOUR
    assert zone.get_dst_savings() == 0
    obj = Formattable(zone)
    tz = TimeZone.create_time_zone("America/Los_Angeles")
    date = 1358208000000.0  # 2013-01-15T00:00:00Z

    # [1]
    # UnicodeString &icu::TimeZoneFormat::format(
    #       const Formattable &obj,
    #       UnicodeString &appendTo,
    #       FieldPosition &pos,
    #       UErrorCode &status
    # )
    append_to = UnicodeString()
    pos = FieldPosition(FieldPosition.DONT_CARE)
    result = fmt.format(obj, append_to, pos)
    assert isinstance(result, UnicodeString)
    assert id(result) == id(append_to)
    assert result == "GMT+09:00"

    # [3]
    # UnicodeString &icu::TimeZoneFormat::format(
    #       const Formattable &obj,
    #       UnicodeString &appendTo,
    #       FieldPositionIterator *posIter,
    #       UErrorCode &status
    # )
    append_to.remove()
    pos_iter = FieldPositionIterator()
    with pytest.raises(ICUError) as exc_info:
        _ = fmt.format(obj, append_to, pos_iter)
    assert exc_info.value.args[0] == UErrorCode.U_UNSUPPORTED_ERROR

    # [4]
    # UnicodeString &icu::TimeZoneFormat::format(
    #       const Formattable &obj,
    #       UnicodeString &appendTo,
    #       UErrorCode &status
    # )
    append_to.remove()
    result = fmt.format(obj, append_to)
    assert isinstance(result, UnicodeString)
    assert id(result) == id(append_to)
    assert result == "GMT+09:00"

    # [5]
    # UnicodeString &icu::TimeZoneFormat::format(
    #       UTimeZoneFormatStyle style,
    #       const TimeZone &tz,
    #       UDate date,
    #       UnicodeString &name,
    #       UTimeZoneFormatTimeType *timeType = NULL
    # )
    name = UnicodeString()
    time_type = UTimeZoneFormatTimeType(0)
    result = fmt.format(
        UTimeZoneFormatStyle.UTZFMT_STYLE_SPECIFIC_LONG,
        tz,
        date,
        name,
        time_type)
    assert isinstance(result, UnicodeString)
    assert id(result) == id(name)
    assert result == "Pacific Standard Time"
    assert time_type == UTimeZoneFormatTimeType.UTZFMT_TIME_TYPE_STANDARD

    name.remove()
    result = fmt.format(
        UTimeZoneFormatStyle.UTZFMT_STYLE_SPECIFIC_LONG,
        tz,
        date,
        name,
        None)
    assert isinstance(result, UnicodeString)
    assert id(result) == id(name)
    assert result == "Pacific Standard Time"

    name.remove()
    result = fmt.format(
        UTimeZoneFormatStyle.UTZFMT_STYLE_SPECIFIC_LONG,
        tz,
        date,
        name)
    assert isinstance(result, UnicodeString)
    assert id(result) == id(name)
    assert result == "Pacific Standard Time"


@pytest.mark.skipif(U_ICU_VERSION_MAJOR_NUM < 51, reason="ICU4C<51")
def test_icu_51():
    fmt = TimeZoneFormat.create_instance("en")
    offset = int(-8.5 * HOUR)
    result = UnicodeString()

    # UnicodeString &icu::TimeZoneFormat::formatOffsetISO8601Basic(
    #       int32_t offset,
    #       UBool useUtcIndicator,
    #       UBool isShort,
    #       UBool ignoreSeconds,
    #       UnicodeString &result,
    #       UErrorCode &status
    # )
    string = fmt.format_offset_iso8601_basic(
        offset,
        True,
        False,
        True,
        result)
    assert isinstance(string, UnicodeString)
    assert id(result) == id(string)
    assert result == "-0830"

    # UnicodeString &icu::TimeZoneFormat::formatOffsetISO8601Extended(
    #       int32_t offset,
    #       UBool useUtcIndicator,
    #       UBool isShort,
    #       UBool ignoreSeconds,
    #       UnicodeString &result,
    #       UErrorCode &status
    # )
    string = fmt.format_offset_iso8601_extended(
        offset,
        True,
        False,
        True,
        result)
    assert isinstance(string, UnicodeString)
    assert id(result) == id(string)
    assert result == "-08:30"

    # UnicodeString &icu::TimeZoneFormat::formatOffsetShortLocalizedGMT(
    #       int32_t offset,
    #       UnicodeString &result,
    #       UErrorCode &status
    # )
    string = fmt.format_offset_short_localized_gmt(
        offset,
        result)
    assert isinstance(string, UnicodeString)
    assert id(result) == id(string)
    assert result == "GMT-8:30"

    # int32_t icu::TimeZoneFormat::parseOffsetShortLocalizedGMT(
    #       const UnicodeString &text,
    #       ParsePosition &pos
    # )
    pos = ParsePosition(0)
    result = fmt.parse_offset_short_localized_gmt(
        UnicodeString("GMT-8:30"),
        pos)
    assert pos.get_error_index() == -1
    assert result == offset

    pos = ParsePosition(0)
    result = fmt.parse_offset_short_localized_gmt(
        "GMT-8:30",
        pos)
    assert pos.get_error_index() == -1
    assert result == offset


def test_parse():
    fmt = TimeZoneFormat.create_instance("en")
    tzid = UnicodeString()

    # [1]
    # TimeZone *icu::TimeZoneFormat::parse(
    #       UTimeZoneFormatStyle style,
    #       const UnicodeString &text,
    #       ParsePosition &pos,
    #       int32_t parseOptions,
    #       UTimeZoneFormatTimeType *timeType = NULL
    # )
    pos = ParsePosition(0)
    time_type = UTimeZoneFormatTimeType(0)
    tz = fmt.parse(
        UTimeZoneFormatStyle.UTZFMT_STYLE_SPECIFIC_SHORT,
        UnicodeString("EST"),
        pos,
        UTimeZoneFormatParseOption.UTZFMT_PARSE_OPTION_ALL_STYLES,
        time_type)
    assert pos.get_error_index() == -1
    assert isinstance(tz, TimeZone)
    assert tz.get_id(tzid) == "America/New_York"
    assert tz.get_raw_offset() == -5 * HOUR
    assert tz.get_dst_savings() == 1 * HOUR
    assert time_type == UTimeZoneFormatTimeType.UTZFMT_TIME_TYPE_STANDARD

    pos = ParsePosition(0)
    tz = fmt.parse(
        UTimeZoneFormatStyle.UTZFMT_STYLE_SPECIFIC_SHORT,
        UnicodeString("EST"),
        pos,
        UTimeZoneFormatParseOption.UTZFMT_PARSE_OPTION_ALL_STYLES,
        None)
    assert pos.get_error_index() == -1
    assert isinstance(tz, TimeZone)
    assert tz.get_id(tzid) == "America/New_York"
    assert tz.get_raw_offset() == -5 * HOUR
    assert tz.get_dst_savings() == 1 * HOUR

    pos = ParsePosition(0)
    tz = fmt.parse(
        UTimeZoneFormatStyle.UTZFMT_STYLE_SPECIFIC_SHORT,
        UnicodeString("EST"),
        pos,
        UTimeZoneFormatParseOption.UTZFMT_PARSE_OPTION_ALL_STYLES)
    assert pos.get_error_index() == -1
    assert isinstance(tz, TimeZone)
    assert tz.get_id(tzid) == "America/New_York"
    assert tz.get_raw_offset() == -5 * HOUR
    assert tz.get_dst_savings() == 1 * HOUR

    pos = ParsePosition(0)
    tz = fmt.parse(
        UTimeZoneFormatStyle.UTZFMT_STYLE_SPECIFIC_SHORT,
        "EST",
        pos,
        UTimeZoneFormatParseOption.UTZFMT_PARSE_OPTION_ALL_STYLES)
    assert pos.get_error_index() == -1
    assert isinstance(tz, TimeZone)
    assert tz.get_id(tzid) == "America/New_York"
    assert tz.get_raw_offset() == -5 * HOUR
    assert tz.get_dst_savings() == 1 * HOUR

    pos = ParsePosition(0)
    tz = fmt.parse(
        UTimeZoneFormatStyle.UTZFMT_STYLE_SPECIFIC_SHORT,
        "AQTST",
        pos,
        UTimeZoneFormatParseOption.UTZFMT_PARSE_OPTION_ALL_STYLES)
    assert pos.get_error_index() != -1
    assert tz is None

    # [2]
    # TimeZone *icu::TimeZoneFormat::parse(
    #       UTimeZoneFormatStyle style,
    #       const UnicodeString &text,
    #       ParsePosition &pos,
    #       UTimeZoneFormatTimeType *timeType = NULL
    # )
    pos = ParsePosition(0)
    time_type = UTimeZoneFormatTimeType(0)
    tz = fmt.parse(
        UTimeZoneFormatStyle.UTZFMT_STYLE_SPECIFIC_SHORT,
        UnicodeString("CST"),
        pos,
        time_type)
    assert pos.get_error_index() == -1
    assert isinstance(tz, TimeZone)
    assert tz.get_id(tzid) == "America/Chicago"
    assert tz.get_raw_offset() == -6 * HOUR
    assert tz.get_dst_savings() == 1 * HOUR
    assert time_type == UTimeZoneFormatTimeType.UTZFMT_TIME_TYPE_STANDARD

    pos = ParsePosition(0)
    tz = fmt.parse(
        UTimeZoneFormatStyle.UTZFMT_STYLE_SPECIFIC_SHORT,
        UnicodeString("CST"),
        pos,
        None)
    assert pos.get_error_index() == -1
    assert isinstance(tz, TimeZone)
    assert tz.get_id(tzid) == "America/Chicago"
    assert tz.get_raw_offset() == -6 * HOUR
    assert tz.get_dst_savings() == 1 * HOUR

    pos = ParsePosition(0)
    tz = fmt.parse(
        UTimeZoneFormatStyle.UTZFMT_STYLE_SPECIFIC_SHORT,
        UnicodeString("CST"),
        pos)
    assert pos.get_error_index() == -1
    assert isinstance(tz, TimeZone)
    assert tz.get_id(tzid) == "America/Chicago"
    assert tz.get_raw_offset() == -6 * HOUR
    assert tz.get_dst_savings() == 1 * HOUR

    pos = ParsePosition(0)
    tz = fmt.parse(
        UTimeZoneFormatStyle.UTZFMT_STYLE_SPECIFIC_SHORT,
        "CST",
        pos)
    assert pos.get_error_index() == -1
    assert isinstance(tz, TimeZone)
    assert tz.get_id(tzid) == "America/Chicago"
    assert tz.get_raw_offset() == -6 * HOUR
    assert tz.get_dst_savings() == 1 * HOUR

    pos = ParsePosition(0)
    tz = fmt.parse(
        UTimeZoneFormatStyle.UTZFMT_STYLE_SPECIFIC_SHORT,
        "AQTST",
        pos)
    assert pos.get_error_index() != -1
    assert tz is None


def test_parse_upcasting():
    from icupy import BasicTimeZone, SimpleTimeZone

    fmt = TimeZoneFormat.create_instance("en")

    pos = ParsePosition(0)
    zone = fmt.parse(
        UTimeZoneFormatStyle.UTZFMT_STYLE_LOCALIZED_GMT,
        "GMT+900",
        pos,
        UTimeZoneFormatParseOption.UTZFMT_PARSE_OPTION_NONE)
    assert pos.get_index() != 0
    assert isinstance(zone, SimpleTimeZone)

    pos = ParsePosition(0)
    zone = fmt.parse(
        UTimeZoneFormatStyle.UTZFMT_STYLE_LOCALIZED_GMT,
        "GMT+900",
        pos)
    assert pos.get_index() != 0
    assert isinstance(zone, SimpleTimeZone)

    # TimeZone -> BasicTimeZone
    pos = ParsePosition(0)
    zone = fmt.parse(
        UTimeZoneFormatStyle.UTZFMT_STYLE_LOCALIZED_GMT,
        "GMT",
        pos,
        UTimeZoneFormatParseOption.UTZFMT_PARSE_OPTION_NONE)
    assert pos.get_index() != 0
    assert not isinstance(zone, SimpleTimeZone)
    assert isinstance(zone, BasicTimeZone)

    pos = ParsePosition(0)
    zone = fmt.parse(
        UTimeZoneFormatStyle.UTZFMT_STYLE_LOCALIZED_GMT,
        "GMT",
        pos)
    assert pos.get_index() != 0
    assert not isinstance(zone, SimpleTimeZone)
    assert isinstance(zone, BasicTimeZone)

    # NULL
    pos = ParsePosition(0)
    zone = fmt.parse(
        UTimeZoneFormatStyle.UTZFMT_STYLE_LOCALIZED_GMT,
        "EST",
        pos)
    assert pos.get_index() == 0
    assert zone is None


def test_parse_object():
    fmt = TimeZoneFormat.create_instance("en")

    # void icu::TimeZoneFormat::parseObject(
    #       const UnicodeString &source,
    #       Formattable &result,
    #       ParsePosition &parse_pos
    # )
    result = Formattable()
    parse_pos = ParsePosition(0)
    fmt.parse_object(UnicodeString("America/Los_Angeles"),
                     result,
                     parse_pos)
    assert parse_pos.get_error_index() == -1
    assert result.get_type() == Formattable.OBJECT
    tz = result.get_object()
    assert isinstance(tz, TimeZone)
    assert tz.get_raw_offset() == -8 * HOUR
    assert tz.get_dst_savings() == 1 * HOUR

    result = Formattable()
    parse_pos = ParsePosition(0)
    fmt.parse_object("America/Los_Angeles",
                     result,
                     parse_pos)
    assert parse_pos.get_error_index() == -1
    assert result.get_type() == Formattable.OBJECT
    tz = result.get_object()
    assert isinstance(tz, TimeZone)
    assert tz.get_raw_offset() == -8 * HOUR
    assert tz.get_dst_savings() == 1 * HOUR

    # [2]
    # void icu::Format::parseObject(
    #       const UnicodeString &source,
    #       Formattable &result,
    #       UErrorCode &status
    # )
    result = Formattable()
    fmt.parse_object(UnicodeString("Asia/Tokyo"), result)
    assert result.get_type() == Formattable.OBJECT
    tz = result.get_object()
    assert isinstance(tz, TimeZone)
    assert tz.get_raw_offset() == 9 * HOUR
    assert tz.get_dst_savings() == 0

    result = Formattable()
    fmt.parse_object("Asia/Tokyo", result)
    assert result.get_type() == Formattable.OBJECT
    tz = result.get_object()
    assert isinstance(tz, TimeZone)
    assert tz.get_raw_offset() == 9 * HOUR
    assert tz.get_dst_savings() == 0



"""
Python bindings for ICU4C
"""

from .icu import *  # noqa

__version__ = VERSION_INFO  # noqa

# DEPRECATED: ICUException is deprecated. Use ICUError instead.
# TODO: Remove ICUException in a future release.
ICUException = ICUError  # noqa



import copy

import pytest
from icupy import (
    ICUError, INT32_MAX, ParsePosition, UErrorCode, UMatchDegree,
    UProperty, USET_IGNORE_SPACE, USetSpanCondition, U_ICU_VERSION_MAJOR_NUM,
    UnicodeSet, UnicodeString,
    u_get_binary_property_set,
)


def test_add():
    test1 = UnicodeSet()
    assert test1.size() == 0

    # [1]
    # UnicodeSet &icu::UnicodeSet::add(const UnicodeString &s)
    #
    # [2]
    # UnicodeSet &icu::UnicodeSet::add(UChar32 c)
    #
    # [3]
    # UnicodeSet &icu::UnicodeSet::add(UChar32 start, UChar32 end)
    result = test1.add(UnicodeString("ab")).add(0xdf).add(0x30, 0x39)
    assert isinstance(result, UnicodeSet)
    assert id(result) == id(test1)
    assert test1.size() == 12  # [0-9\u00DF{ab}]

    result = test1.add("cd").add("ef")
    assert isinstance(result, UnicodeSet)
    assert id(result) == id(test1)
    assert test1.size() == 14  # [0-9ab\u00DF{ab}{cd}{ef}]


def test_add_all():
    test1 = UnicodeSet()
    assert test1.size() == 0
    test2 = UnicodeSet(UnicodeString("[\u00DF{ab}]"))
    assert test2.size() == 2

    # [1]
    # UnicodeSet &icu::UnicodeSet::addAll(const UnicodeSet &c)
    #
    # [2]
    # UnicodeSet &icu::UnicodeSet::addAll(const UnicodeString &s)
    result = test1.add_all(test2).add_all(UnicodeString("0123456789"))
    assert isinstance(result, UnicodeSet)
    assert id(result) == id(test1)
    assert test1.size() == 2 + 10  # [0-9\u00DF{ab}]

    result = test1.add_all("cd").add_all("ef")
    assert isinstance(result, UnicodeSet)
    assert id(result) == id(test1)
    assert test1.size() == 2 + 10 + 4  # [0-9c-f\u00DF{ab}]


def test_add_match_set_to():
    test1 = UnicodeSet(UnicodeString("[0-9]"))
    test2 = UnicodeSet(UnicodeString("[a-z]"))
    assert test1.size() == 10
    assert test2.size() == 26

    # void icu::UnicodeSet::addMatchSetTo(UnicodeSet &toUnionTo)
    test1.add_match_set_to(test2)
    assert test1.size() == 10  # [0-9]
    assert test1.contains(0x30, 0x39)
    assert test2.size() == 10 + 26  # [0-9a-z]
    assert test2.contains(0x30, 0x39)
    assert test2.contains(0x61, 0x7a)


def test_api():
    test1 = UnicodeSet(0x30, 0x39)

    # int32_t icu::UnicodeSet::getRangeCount(void)
    assert test1.get_range_count() == 1

    # UChar32 icu::UnicodeSet::getRangeEnd(int32_t index)
    assert test1.get_range_end(0) == 0x39

    # UChar32 icu::UnicodeSet::getRangeStart(int32_t index)
    assert test1.get_range_start(0) == 0x30

    # int32_t icu::UnicodeSet::indexOf(UChar32 c)
    assert test1.index_of(0x39) == 9
    assert test1.index_of(0x61) == -1

    # UBool icu::UnicodeSet::isBogus(void)
    assert not test1.is_bogus()

    # UBool icu::UnicodeSet::isEmpty(void)
    assert not test1.is_empty()

    # UBool icu::UnicodeSet::isFrozen()
    assert not test1.is_frozen()

    # int32_t icu::UnicodeSet::size(void)
    assert test1.size() == 10
    assert len(test1) == 10

    # UnicodeSet &icu::UnicodeSet::compact()
    result = test1.compact()
    assert isinstance(result, UnicodeSet)
    assert id(result) == id(test1)

    # UnicodeSet &icu::UnicodeSet::clear(void)
    result = test1.clear()
    assert isinstance(result, UnicodeSet)
    assert id(result) == id(test1)

    assert not test1.is_bogus()
    assert test1.is_empty()
    assert not test1.is_frozen()
    assert test1.size() == 0
    assert len(test1) == 0

    # UnicodeSet *icu::UnicodeSet::freeze()
    test1.add(0x30, 0x39)
    result = test1.freeze()
    assert isinstance(result, UnicodeSet)
    assert id(result) == id(test1)
    assert not test1.is_bogus()
    assert not test1.is_empty()
    assert test1.is_frozen()
    assert test1.size() == 10
    assert len(test1) == 10

    # static UBool icu::UnicodeSet::resemblesPattern(
    #       const UnicodeString &pattern,
    #       int32_t pos
    # )
    pattern = UnicodeString("[0-9\u00DF{ab}]")
    assert UnicodeSet.resembles_pattern(pattern, 0)
    assert not UnicodeSet.resembles_pattern(pattern, 1)

    assert UnicodeSet.resembles_pattern("[0-9\u00DF{ab}]", 0)
    assert not UnicodeSet.resembles_pattern("[0-9\u00DF{ab}]", 1)

    test2 = UnicodeSet(UnicodeString("[0-9\u00DF{ab}]"))
    assert not test2.is_bogus()
    assert test2.size() == 12

    # UnicodeString &icu::UnicodeSet::toPattern(
    #       UnicodeString &result,
    #       UBool escapeUnprintable = false
    # )
    result = UnicodeString()
    output = test2.to_pattern(result)
    assert isinstance(output, UnicodeString)
    assert id(result) == id(output)
    assert result == "[0-9\xDF{ab}]"

    output = test2.to_pattern(result, False)
    assert isinstance(output, UnicodeString)
    assert id(result) == id(output)
    assert result == "[0-9\xDF{ab}]"

    output = test2.to_pattern(result, True)
    assert isinstance(output, UnicodeString)
    assert id(result) == id(output)
    assert result == "[0-9\\u00DF{ab}]"

    # UnicodeSet &icu::UnicodeSet::set(
    #       UChar32 start,
    #       UChar32 end
    # )
    result = test2.set(0x30, 0x39)
    assert isinstance(result, UnicodeSet)
    assert id(result) == id(test2)
    assert test2.size() == 10

    # void icu::UnicodeSet::setToBogus()
    test2.set_to_bogus()
    assert test2.is_bogus()


def test_apply_int_property_value():
    test1 = UnicodeSet()
    assert test1.size() == 0

    # UnicodeSet &icu::UnicodeSet::applyIntPropertyValue(
    #       UProperty prop,
    #       int32_t value,
    #       UErrorCode &ec
    # )
    result = test1.apply_int_property_value(
        UProperty.UCHAR_CANONICAL_COMBINING_CLASS,
        8)  # 8 = Kana Voicing
    assert isinstance(result, UnicodeSet)
    assert id(result) == id(test1)
    assert test1.size() == 2  # [\u3099\u309A]
    assert test1.contains(0x3099)
    assert test1.contains(0x309a)

    with pytest.raises(ICUError) as exc_info:
        test1.apply_int_property_value(UProperty.UCHAR_INT_LIMIT, 0)
    assert exc_info.value.args[0] == UErrorCode.U_ILLEGAL_ARGUMENT_ERROR


def test_apply_pattern():
    # [1]
    # UnicodeSet &icu::UnicodeSet::applyPattern(
    #       const UnicodeString &pattern,
    #       ParsePosition &pos,
    #       uint32_t options,
    #       const SymbolTable *symbols,
    #       UErrorCode &status
    # )
    test1 = UnicodeSet()
    pattern = UnicodeString("[a-z{ab}]")
    pos = ParsePosition()
    options = UnicodeSet.IGNORE_SPACE | UnicodeSet.CASE_INSENSITIVE
    result = test1.apply_pattern(pattern, pos, options, None)
    assert isinstance(result, UnicodeSet)
    assert id(result) == id(test1)
    assert pos.get_index() == pattern.length()
    assert test1.size() == 26 * 2 + 3  # [A-Za-z\u017F\u212A{ab}]
    assert test1.contains(0x41, 0x5a)
    assert test1.contains(0x61, 0x7a)

    pos = ParsePosition()
    result = test1.apply_pattern("[0-9]", pos, options, None)
    assert isinstance(result, UnicodeSet)
    assert id(result) == id(test1)
    assert pos.get_index() == 5
    assert test1.size() == 10  # [0-9]
    assert test1.contains(0x30, 0x39)

    with pytest.raises(ICUError) as exc_info:
        test1.apply_pattern(UnicodeString(), pos, options, None)
    assert exc_info.value.args[0] == UErrorCode.U_MALFORMED_SET

    # [2]
    # UnicodeSet &icu::UnicodeSet::applyPattern(
    #       const UnicodeString &pattern,
    #       UErrorCode &status
    # )
    test2 = UnicodeSet()
    result = test2.apply_pattern(pattern)
    assert isinstance(result, UnicodeSet)
    assert id(result) == id(test2)
    assert test2.size() == 26 + 1  # [a-z{ab}]
    assert test2.contains(0x61, 0x7a)

    result = test2.apply_pattern("[0-9]")
    assert isinstance(result, UnicodeSet)
    assert id(result) == id(test2)
    assert test2.size() == 10  # [0-9]
    assert test2.contains(0x30, 0x39)

    with pytest.raises(ICUError) as exc_info:
        test2.apply_pattern(UnicodeString())
    assert exc_info.value.args[0] == UErrorCode.U_MALFORMED_SET


def test_apply_property_alias():
    test1 = UnicodeSet()

    # UnicodeSet &icu::UnicodeSet::applyPropertyAlias(
    #       const UnicodeString &prop,
    #       const UnicodeString &value,
    #       UErrorCode &ec
    # )
    result = test1.apply_property_alias(
        UnicodeString("Canonical_Combining_Class"),
        UnicodeString("Kana_Voicing"))
    assert isinstance(result, UnicodeSet)
    assert id(result) == id(test1)
    assert test1.size() == 2  # [\u3099\u309A]
    assert test1.contains(0x3099)
    assert test1.contains(0x309a)

    with pytest.raises(ICUError) as exc_info:
        test1.apply_property_alias(UnicodeString("ccc"),
                                   UnicodeString("-"))
    assert exc_info.value.args[0] == UErrorCode.U_ILLEGAL_ARGUMENT_ERROR

    result = test1.apply_property_alias("ccc", UnicodeString("10"))
    assert isinstance(result, UnicodeSet)
    assert id(result) == id(test1)
    assert test1.size() == 1
    assert test1.contains(0x05B0)

    result = test1.apply_property_alias(UnicodeString("ccc"), "11")
    assert isinstance(result, UnicodeSet)
    assert id(result) == id(test1)
    assert test1.size() == 1
    assert test1.contains(0x05B1)

    result = test1.apply_property_alias("ccc", "12")
    assert isinstance(result, UnicodeSet)
    assert id(result) == id(test1)
    assert test1.size() == 1
    assert test1.contains(0x05B2)


def test_char_at():
    test1 = UnicodeSet(0x30, 0x39)
    assert test1.size() == 10

    # UChar32 icu::UnicodeSet::charAt(int32_t index)
    assert test1.char_at(9) == 0x39
    assert test1.char_at(10) == -1
    assert test1.char_at(-1) == -1

    # __getitem__(self, index: int) -> int
    assert test1[9] == 0x39
    with pytest.raises(IndexError):  # IndexError: list index out of range
        _ = test1[10]
    assert test1[-1] == 0x39

    # __getitem__(self, slice: slice) -> List[int]
    assert test1[1:4] == [0x31, 0x32, 0x33]
    assert test1[1:6:2] == [0x31, 0x33, 0x35]
    with pytest.raises(ValueError):  # ValueError: slice step cannot be zero
        _ = test1[::0]

    test1.add(UnicodeString("ab"))
    t = [c for c in test1]
    assert t == [
        0x30, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37, 0x38, 0x39, -1,
    ]


def test_clone():
    test1 = UnicodeSet(0x30, 0x39)
    assert not test1.is_frozen()

    # UnicodeSet *icu::UnicodeSet::clone()
    test2 = test1.clone()
    assert isinstance(test2, UnicodeSet)
    assert test2 == test1
    assert not test2.is_frozen()

    test1.freeze()
    test3 = test1.clone()
    assert test3 == test1
    assert test1.is_frozen()
    assert test3.is_frozen()

    # UnicodeSet *icu::UnicodeSet::cloneAsThawed()
    test4 = test1.clone_as_thawed()
    assert isinstance(test4, UnicodeSet)
    assert test4 == test1
    assert test1.is_frozen()
    assert not test4.is_frozen()

    test5 = copy.copy(test1)
    assert test5 == test1
    assert test5.is_frozen()

    test6 = copy.deepcopy(test1)
    assert test6 == test1
    assert test6.is_frozen()


def test_close_over():
    pattern = UnicodeString("[aq\u00DF{Bc}{bC}{Fi}]")
    test1 = UnicodeSet(pattern)
    test2 = test1.clone()

    # UnicodeSet &icu::UnicodeSet::closeOver(int32_t attribute)
    result = test2.close_over(UnicodeSet.CASE_INSENSITIVE |
                              UnicodeSet.ADD_CASE_MAPPINGS)
    assert isinstance(result, UnicodeSet)
    assert id(result) == id(test2)


def test_complement():
    test1 = UnicodeSet(UnicodeSet.MIN_VALUE, UnicodeSet.MAX_VALUE)

    # [1]
    # UnicodeSet &icu::UnicodeSet::complement(const UnicodeString &s)
    #
    # [2]
    # UnicodeSet &icu::UnicodeSet::complement(UChar32 c)
    #
    # [3]
    # UnicodeSet &icu::UnicodeSet::complement(UChar32 start, UChar32 end)
    #
    # [4]
    # UnicodeSet &icu::UnicodeSet::complement()
    result = (test1.complement(UnicodeString("ab"))
              .complement("cd")
              .complement(0xdf)
              .complement(0x30, 0x39)
              .complement())
    assert isinstance(result, UnicodeSet)
    assert id(result) == id(test1)
    assert test1.size() == 13  # [0-9\u00DF{ab}{cd}]
    assert test1.contains(0x30, 0x39)
    assert test1.contains(0xdf)
    assert test1.contains("ab")
    assert test1.contains("cd")


def test_complement_all():
    test1 = UnicodeSet(UnicodeString("[0-9\u00DF{ab}]"))
    assert test1.size() == 12
    test2 = UnicodeSet(0x30, 0x39)
    assert test2.size() == 10

    # [1]
    # UnicodeSet &icu::UnicodeSet::complementAll(const UnicodeSet &c)
    #
    # [2]
    # UnicodeSet &icu::UnicodeSet::complementAll(const UnicodeString &s)
    result = (test1.complement_all(test2)
              .complement_all(UnicodeString("ab"))
              .complement_all("cd"))
    assert isinstance(result, UnicodeSet)
    assert id(result) == id(test1)
    assert test1.size() == 12 - 10 + 2 + 2  # [a-d\u00DF{ab}]
    assert test1.contains(0x61, 0x64)
    assert test1.contains(0xdf)
    assert test1.contains("ab")


def test_contains():
    test1 = UnicodeSet(0x30, 0x39)

    # [1]
    # UBool icu::UnicodeSet::contains(const UnicodeString &s)
    assert test1.contains(UnicodeString("0"))
    assert not test1.contains(UnicodeString("01"))

    assert test1.contains("0")
    assert not test1.contains("01")

    assert UnicodeString("0") in test1
    assert UnicodeString("01") not in test1

    assert "0" in test1
    assert "01" not in test1

    # [2]
    # UBool icu::UnicodeSet::contains(UChar32 c)
    assert test1.contains(0x30)
    assert not test1.contains(0x61)

    assert 0x30 in test1
    assert 0x61 not in test1

    # [3]
    # UBool icu::UnicodeSet::contains(UChar32 start, UChar32 end)
    assert test1.contains(0x30, 0x39)
    assert not test1.contains(0x61, 0x61)


def test_contains_all():
    test1 = UnicodeSet(0x30, 0x39)

    # [1]
    # UBool icu::UnicodeSet::containsAll(const UnicodeSet &c)
    assert test1.contains_all(UnicodeSet(0x30, 0x39))
    assert not test1.contains_all(UnicodeSet(0x2f, 0x3a))

    # [2]
    # UBool icu::UnicodeSet::containsAll(const UnicodeString &s)
    assert test1.contains_all(UnicodeString("0123456789"))
    assert not test1.contains_all(UnicodeString("/0123456789:"))

    assert test1.contains_all("0123456789")
    assert not test1.contains_all("/0123456789:")


def test_contains_none():
    test1 = UnicodeSet(0x30, 0x39)

    # [1]
    # UBool icu::UnicodeSet::containsNone(const UnicodeSet &c)
    assert not test1.contains_none(UnicodeSet(0x2f, 0x3a))
    assert test1.contains_none(UnicodeSet(0x2f, 0x2f))

    # [2]
    # UBool icu::UnicodeSet::containsNone(const UnicodeString &s)
    assert not test1.contains_none(UnicodeString("/0123456789:"))
    assert test1.contains_none(UnicodeString("/"))

    assert not test1.contains_none("/0123456789:")
    assert test1.contains_none("/")

    # [3]
    # UBool icu::UnicodeSet::containsNone(UChar32 start, UChar32 end)
    assert not test1.contains_none(0x2f, 0x3a)
    assert test1.contains_none(0x2f, 0x2f)


def test_contains_some():
    test1 = UnicodeSet(0x30, 0x39)

    # [1]
    # UBool icu::UnicodeSet::containsSome(const UnicodeSet &s)
    assert test1.contains_some(UnicodeSet(UnicodeString("[0-9{ab}]")))
    assert not test1.contains_some(UnicodeSet(UnicodeString("[a-z{ab}]")))

    # [2]
    # UBool icu::UnicodeSet::containsSome(const UnicodeString &s)
    assert test1.contains_some(UnicodeString("ab0123456789"))
    assert not test1.contains_some(UnicodeString("ab"))

    assert test1.contains_some("ab0123456789")
    assert not test1.contains_some("ab")

    # [3]
    # UBool icu::UnicodeSet::containsSome(UChar32 start, UChar32 end)
    assert test1.contains_some(0x2f, 0x3a)
    assert not test1.contains_some(0x2f, 0x2f)


def test_create_from():
    # static UnicodeSet *icu::UnicodeSet::createFrom(const UnicodeString &s)
    test1 = UnicodeSet.create_from(UnicodeString("ab"))
    assert isinstance(test1, UnicodeSet)
    assert test1.size() == 1  # [{ab}]
    assert test1.contains(UnicodeString("ab"))

    test1a = UnicodeSet.create_from("ab")
    assert isinstance(test1a, UnicodeSet)
    assert test1a.size() == 1  # [{ab}]
    assert test1a.contains(UnicodeString("ab"))


def test_create_from_all():
    # static UnicodeSet *icu::UnicodeSet::createFromAll(const UnicodeString &s)
    test1 = UnicodeSet.create_from_all(UnicodeString("ab"))
    assert isinstance(test1, UnicodeSet)
    assert test1.size() == 2  # [ab]
    assert test1.contains(0x61, 0x62)

    test1 = UnicodeSet.create_from_all("ab")
    assert isinstance(test1, UnicodeSet)
    assert test1.size() == 2  # [ab]
    assert test1.contains(0x61, 0x62)


def test_from_uset():
    # [1]
    # const UnicodeSet *icu::UnicodeSet::fromUSet(const USet *uset)
    uset1 = u_get_binary_property_set(UProperty.UCHAR_POSIX_BLANK)
    test1 = UnicodeSet.from_uset(uset1)
    assert isinstance(test1, UnicodeSet)
    assert test1.size() > 0
    assert test1.contains(0x20)  # U+0020: Space (SP)

    # USet *icu::UnicodeSet::toUSet()
    uset2 = test1.to_uset()

    # [2]
    # UnicodeSet *icu::UnicodeSet::fromUSet(USet *uset)
    test2 = UnicodeSet.from_uset(uset2)
    assert isinstance(test2, UnicodeSet)
    assert test2 == test1


def test_hash_code():
    test1 = UnicodeSet(0x30, 0x39)
    test2 = UnicodeSet()

    # int32_t icu::UnicodeSet::hashCode(void)
    assert test1.hash_code() != test2.hash_code()

    test2.add(0x30, 0x39)
    assert test1.hash_code() == test2.hash_code()


def test_matches():
    test1 = UnicodeSet(UnicodeString("[0-9{abc}]"))
    text = UnicodeString("abcd 789")

    # UMatchDegree icu::UnicodeSet::matches(
    #       const Replaceable &text,
    #       int32_t &offset,
    #       int32_t limit,
    #       UBool incremental
    # )
    offset = 0
    result, offset = test1.matches(text, offset, text.length(), True)
    assert result == UMatchDegree.U_MATCH
    assert offset == 3

    result, offset = test1.matches(text, offset, text.length(), True)
    assert result == UMatchDegree.U_MISMATCH
    assert offset == 3

    offset = 5
    result, offset = test1.matches(text, offset, text.length(), True)
    assert result == UMatchDegree.U_MATCH
    assert offset == 6


def test_operator():
    test1 = UnicodeSet(0x30, 0x39)
    test2 = UnicodeSet(UnicodeString("[0-9]"))
    test3 = UnicodeSet(UnicodeString("[0-9{ab}]"))

    # UBool icu::UnicodeSet::operator!=(const UnicodeSet &o)
    assert not (test1 != test2)
    assert test1 != test3
    assert test2 != test3

    # UBool icu::UnicodeSet::operator==(const UnicodeSet &o)
    assert test1 == test2
    assert not (test1 == test3)
    assert not (test2 == test3)


def test_remove():
    test1 = UnicodeSet(UnicodeString("[0-9\u00DF{ab}]"))
    assert test1.size() == 12

    # [1]
    # UnicodeSet &icu::UnicodeSet::remove(const UnicodeString &s)
    #
    # [2]
    # UnicodeSet &icu::UnicodeSet::remove(UChar32 c)
    #
    # [3]
    # UnicodeSet &icu::UnicodeSet::remove(UChar32 start, UChar32 end)
    result = test1.remove(UnicodeString("ab")).remove(0xdf).remove(0x30, 0x39)
    assert isinstance(result, UnicodeSet)
    assert id(result) == id(test1)
    assert test1.size() == 0

    test1 = UnicodeSet(UnicodeString("[0-9\u00DF{ab}]"))
    result = test1.remove("ab").remove(0xdf).remove(0x30, 0x39)
    assert isinstance(result, UnicodeSet)
    assert id(result) == id(test1)
    assert test1.size() == 0


def test_remove_all():
    test1 = UnicodeSet(UnicodeString("[0-9\u00DF{ab}]"))
    test2 = test1.clone()
    assert test2.size() == 12
    test3 = UnicodeSet(0x30, 0x39)

    # [1]
    # UnicodeSet &icu::UnicodeSet::removeAll(const UnicodeSet &c)
    #
    # [2]
    # UnicodeSet &icu::UnicodeSet::removeAll(const UnicodeString &s)
    result = test2.remove_all(test3).remove_all(UnicodeString("ab"))
    assert isinstance(result, UnicodeSet)
    assert id(result) == id(test2)
    assert test2.size() == 12 - 10  # [\u00DF{ab}]
    assert test2.contains(0xdf)
    assert test2.contains(UnicodeString("ab"))

    test2 = test1.clone()
    result = test2.remove_all(test3).remove_all("ab")
    assert isinstance(result, UnicodeSet)
    assert id(result) == id(test2)
    assert test2.size() == 12 - 10  # [\u00DF{ab}]
    assert test2.contains(0xdf)
    assert test2.contains(UnicodeString("ab"))


def test_remove_all_strings():
    test1 = UnicodeSet(UnicodeString("[0-9\u00DF{ab}]"))
    assert test1.size() == 12

    # UnicodeSet &icu::UnicodeSet::removeAllStrings()
    result = test1.remove_all_strings()
    assert isinstance(result, UnicodeSet)
    assert id(result) == id(test1)
    assert test1.size() == 11  # [0-9\u00DF]
    assert test1.contains(0x30, 0x39)
    assert test1.contains(0xdf)
    assert not test1.contains("ab")


def test_retain():
    test1 = UnicodeSet(UnicodeString("[0-9\u00DF{ab}]"))
    test2 = test1.clone()
    test3 = test1.clone()
    assert test1.size() == 12

    # [1]
    # UnicodeSet &icu::UnicodeSet::retain(UChar32 c)
    result = test2.retain(0xdf)
    assert isinstance(result, UnicodeSet)
    assert id(result) == id(test2)
    assert test2.size() == 1 + 1  # [\u00DF{ab}]
    assert test1.contains(0xdf)

    # [2]
    # UnicodeSet &icu::UnicodeSet::retain(UChar32 start, UChar32 end)
    result = test3.retain(0x30, 0x39)
    assert isinstance(result, UnicodeSet)
    assert id(result) == id(test3)
    assert test3.size() == 10 + 1  # [0-9{ab}]
    assert test1.contains(0x30, 0x39)
    assert test1.contains(UnicodeString("ab"))


@pytest.mark.skipif(U_ICU_VERSION_MAJOR_NUM < 69, reason="ICU4C<69")
def test_retain_69():
    test1 = UnicodeSet("[0-9\u00DF{ab}]")
    test2 = test1.clone()
    test3 = test1.clone()
    assert test1.size() == 12

    # [1]
    # UnicodeSet &icu::UnicodeSet::retain(const UnicodeString &s)
    result = test2.retain(UnicodeString("\xdf"))
    assert isinstance(result, UnicodeSet)
    assert id(result) == id(test2)
    assert test2.size() == 1 + 1  # [\u00DF{ab}]
    assert test2.contains(0xdf)
    assert test2.contains("ab")

    result = test3.retain("ab")
    assert isinstance(result, UnicodeSet)
    assert id(result) == id(test3)
    assert test3.size() == 1  # [{ab}]
    assert test3.contains("ab")


def test_retain_all():
    test1 = UnicodeSet(UnicodeString("[0-9\u00DF{ab}]"))
    test2 = test1.clone()
    test3 = test1.clone()
    assert test1.size() == 12

    # [1]
    # UnicodeSet &icu::UnicodeSet::retainAll(const UnicodeSet &c)
    result = test2.retain_all(UnicodeSet(0xdf, 0xdf))
    assert isinstance(result, UnicodeSet)
    assert id(result) == id(test2)
    assert test2.size() == 1  # [\u00DF]
    assert test2.contains(0xdf)

    # [2]
    # UnicodeSet &icu::UnicodeSet::retainAll(const UnicodeString &s)
    result = test3.retain_all(UnicodeString("ab"))
    assert isinstance(result, UnicodeSet)
    assert id(result) == id(test3)
    assert test3.size() == 0

    test3 = test1.clone()
    result = test3.retain_all("ab")
    assert isinstance(result, UnicodeSet)
    assert id(result) == id(test3)
    assert test3.size() == 0


def test_serialize():
    # int32_t icu::UnicodeSet::serialize(
    #       uint16_t *dest,
    #       int32_t destCapacity,
    #       UErrorCode &ec
    # )
    test1 = UnicodeSet(UnicodeString("[0-9\u00DF{ab}]"))
    dest = test1.serialize()
    assert isinstance(dest, list)
    assert dest == [
        4,  # length=((n+2*m)|(m!=0?0x8000:0)) (n=4, m=0)
        ord("0"), ord("9") + 1,  # bmp code points
        0xdf, 0xdf + 1,
    ]

    test2 = UnicodeSet(UnicodeString("[\U0001F000-\U0001F003]"))
    dest = test2.serialize()
    assert isinstance(dest, list)
    assert dest == [
        (0 + 2 * 2) | 0x8000,  # length=((n+2*m)|(m!=0?0x8000:0)) (n=0, m=2)
        0,  # bmpLength
        1, 0xf000,  # supplementary code points
        (0x1f003 + 1) >> 16, (0x1f003 + 1) & 0xffff,
    ]


def test_span():
    test1 = UnicodeSet(UnicodeString("[0-9\u00DF{ab}]"))

    # [1]
    # int32_t icu::UnicodeSet::span(
    #       const char16_t *s,
    #       int32_t length,
    #       USetSpanCondition spanCondition
    # )
    assert test1.span(
        "abcde",
        -1,
        USetSpanCondition.USET_SPAN_CONTAINED) == 2

    # [2]
    # int32_t icu::UnicodeSet::span(
    #       const UnicodeString &s,
    #       int32_t start,
    #       USetSpanCondition spanCondition
    # )
    assert test1.span(
        UnicodeString("abcde"),
        0,
        USetSpanCondition.USET_SPAN_CONTAINED) == 2


def test_span_back():
    test1 = UnicodeSet(UnicodeString("[0-9\u00DF{ab}]"))

    # [1]
    # int32_t icu::UnicodeSet::spanBack(
    #       const char16_t *s,
    #       int32_t length,
    #       USetSpanCondition spanCondition
    # )
    assert test1.span_back(
        "aaaab",
        -1,
        USetSpanCondition.USET_SPAN_CONTAINED) == 3

    # [2]
    # int32_t icu::UnicodeSet::spanBack(
    #       const UnicodeString &s,
    #       int32_t limit,
    #       USetSpanCondition spanCondition
    # )
    assert test1.span_back(
        UnicodeString("aaaab"),
        INT32_MAX,
        USetSpanCondition.USET_SPAN_CONTAINED) == 3


def test_unicode_set():
    # [1]
    # icu::UnicodeSet::UnicodeSet()
    test1 = UnicodeSet()
    assert test1.size() == 0

    # [2]
    # icu::UnicodeSet::UnicodeSet(UChar32 start, UChar32 end)
    test2 = UnicodeSet(0x30, 0x39)
    assert test2.size() == 10
    assert test2.contains(0x30, 0x39)

    # [4]
    # icu::UnicodeSet::UnicodeSet(
    #       const UnicodeString &pattern,
    #       UErrorCode &status
    # )
    test4 = UnicodeSet(UnicodeString("[0-9]"))
    assert test4.size() == 10
    assert test4.contains(0x30, 0x39)

    test4a = UnicodeSet("[0-9]")
    assert test4a.size() == 10
    assert test4a.contains(0x30, 0x39)

    # [6]
    # icu::UnicodeSet::UnicodeSet(
    #       const UnicodeString &pattern,
    #       ParsePosition &pos,
    #       uint32_t options,
    #       const SymbolTable *symbols,
    #       UErrorCode &status
    # )
    pattern = UnicodeString("foo [ 0-9 ]")
    pos = ParsePosition()
    pos.set_index(4)
    options = USET_IGNORE_SPACE
    test6 = UnicodeSet(pattern, pos, options, None)
    assert pos.get_index() == pattern.length()
    assert test6.size() == 10
    assert test6.contains(0x30, 0x39)

    pos = ParsePosition()
    pos.set_index(4)
    test6a = UnicodeSet("foo [ 0-9 ]", pos, options, None)
    assert pos.get_index() == 11
    assert test6a.size() == 10
    assert test6a.contains(0x30, 0x39)

    # [7]
    # icu::UnicodeSet::UnicodeSet(const UnicodeSet &o)
    test7 = UnicodeSet(test2)
    assert test7.size() == 10
    assert test7.contains(0x30, 0x39)



import copy

import pytest
from icupy import (
    BreakIterator, CharacterIterator, Locale, RuleBasedBreakIterator,
    StringCharacterIterator, ULocDataLocaleType,
    UParseError, UWordBreak, U_ICU_VERSION_MAJOR_NUM, UnicodeString,
    utext_close, utext_extract, utext_native_length,
    utext_open_const_unicode_string,
)


def test_adopt_text():
    bi = BreakIterator.create_line_instance(Locale.get_us())
    src = UnicodeString("foo bar baz.")
    it = StringCharacterIterator(src)

    # void icu::BreakIterator::adoptText(CharacterIterator *it)
    bi.adopt_text(it)
    assert bi.first() == 0
    assert bi.next() == 4
    assert bi.next() == 8
    assert bi.next() == 12
    assert bi.next() == BreakIterator.DONE

    bi.adopt_text(None)
    assert bi.first() == 0
    assert bi.next() == BreakIterator.DONE


def test_clone():
    bi1 = BreakIterator.create_word_instance(Locale.get_us())
    src = UnicodeString("foo bar baz.")
    bi1.set_text(src)
    bi1.next()

    # BreakIterator *icu::BreakIterator::clone()
    bi2 = bi1.clone()
    assert isinstance(bi2, BreakIterator)
    assert id(bi1) != id(bi2)
    assert bi1 == bi2
    assert bi1.current() == bi2.current()

    bi3 = copy.copy(bi1)
    assert bi1 == bi3
    assert bi1.current() == bi3.current()

    bi4 = copy.deepcopy(bi1)
    assert bi1 == bi4
    assert bi1.current() == bi4.current()


def test_create_character_instance():
    # static BreakIterator *icu::BreakIterator::createCharacterInstance(
    #       const Locale &where,
    #       UErrorCode &status
    # )
    bi = BreakIterator.create_character_instance(Locale.get_us())
    assert isinstance(bi, BreakIterator)
    src = UnicodeString("foo bar baz.")
    bi.set_text(src)

    assert bi.first() == 0
    assert bi.next() == 1
    assert bi.next() == 2
    assert bi.next() == 3
    assert bi.next() == 4
    assert bi.next() == 5
    assert bi.next() == 6
    assert bi.next() == 7
    assert bi.next() == 8
    assert bi.next() == 9
    assert bi.next() == 10
    assert bi.next() == 11
    assert bi.next() == 12
    assert bi.next() == BreakIterator.DONE

    assert list(bi) == [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12]
    assert reversed(bi) == [12, 11, 10, 9, 8, 7, 6, 5, 4, 3, 2, 1]

    bi = BreakIterator.create_character_instance("en_US")
    bi.set_text(src)
    assert list(bi) == [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12]
    assert reversed(bi) == [12, 11, 10, 9, 8, 7, 6, 5, 4, 3, 2, 1]

    loc = bi.get_locale(ULocDataLocaleType.ULOC_VALID_LOCALE)
    assert loc.get_name() == "en_US"


def test_create_line_instance():
    # static BreakIterator *icu::BreakIterator::createLineInstance(
    #       const Locale &where,
    #       UErrorCode &status
    # )
    bi = BreakIterator.create_line_instance(Locale.get_us())
    assert isinstance(bi, BreakIterator)
    src = UnicodeString("foo bar baz.")
    bi.set_text(src)

    assert bi.first() == 0
    assert bi.next() == 4
    assert bi.next() == 8
    assert bi.next() == 12
    assert bi.next() == BreakIterator.DONE

    assert list(bi) == [4, 8, 12]
    assert reversed(bi) == [12, 8, 4]

    bi = BreakIterator.create_line_instance("en_US")
    bi.set_text(src)
    assert list(bi) == [4, 8, 12]
    assert reversed(bi) == [12, 8, 4]

    loc = bi.get_locale(ULocDataLocaleType.ULOC_VALID_LOCALE)
    assert loc.get_name() == "en_US"


def test_create_sentence_instance():
    # static BreakIterator *icu::BreakIterator::createSentenceInstance(
    #       const Locale &where,
    #       UErrorCode &status
    # )
    bi = BreakIterator.create_sentence_instance(Locale.get_us())
    assert isinstance(bi, BreakIterator)
    src = UnicodeString("foo bar baz.")
    bi.set_text(src)

    assert bi.first() == 0
    assert bi.next() == 12
    assert bi.next() == BreakIterator.DONE

    assert list(bi) == [12]
    assert reversed(bi) == [12]

    bi = BreakIterator.create_sentence_instance("en_US")
    bi.set_text(src)
    assert list(bi) == [12]
    assert reversed(bi) == [12]

    loc = bi.get_locale(ULocDataLocaleType.ULOC_VALID_LOCALE)
    assert loc.get_name() == "en_US"


def test_create_word_instance():
    # static BreakIterator *icu::BreakIterator::createWordInstance(
    #       const Locale &where,
    #       UErrorCode &status
    # )
    bi = BreakIterator.create_word_instance(Locale.get_us())
    assert isinstance(bi, BreakIterator)
    src = UnicodeString("foo bar baz.")
    bi.set_text(src)

    assert bi.first() == 0
    assert bi.next() == 3
    assert bi.next() == 4
    assert bi.next() == 7
    assert bi.next() == 8
    assert bi.next() == 11
    assert bi.next() == 12
    assert bi.next() == BreakIterator.DONE

    assert list(bi) == [3, 4, 7, 8, 11, 12]
    assert reversed(bi) == [12, 11, 8, 7, 4, 3]

    bi = BreakIterator.create_word_instance("en_US")
    bi.set_text(src)
    assert list(bi) == [3, 4, 7, 8, 11, 12]
    assert reversed(bi) == [12, 11, 8, 7, 4, 3]

    loc = bi.get_locale(ULocDataLocaleType.ULOC_VALID_LOCALE)
    assert loc.get_name() == "en_US"


def test_following():
    bi = BreakIterator.create_word_instance(Locale.get_us())
    src = UnicodeString("foo bar baz.")
    bi.set_text(src)

    # int32_t icu::BreakIterator::following(int32_t offset)
    # [0, 3, 4, 7, 8, 11, 12]
    assert bi.following(0) == 3
    assert bi.following(3) == 4
    assert bi.following(4) == 7
    assert bi.following(7) == 8
    assert bi.following(8) == 11
    assert bi.following(11) == 12
    assert bi.following(12) == BreakIterator.DONE


def test_get_available_locales():
    # [1]
    # static const Locale *icu::BreakIterator::getAvailableLocales(
    #       int32_t &count
    # )
    it = BreakIterator.get_available_locales()
    assert isinstance(it, list)
    assert len(it) > 0
    assert all(isinstance(x, Locale) for x in it)
    assert Locale("ja") in it


def test_get_display_name():
    default_locale = None
    try:
        default_locale = Locale.get_default()
        us_locale = Locale.get_us()
        Locale.set_default(us_locale)
        name = UnicodeString()

        # [1]
        # static UnicodeString &icu::BreakIterator::getDisplayName(
        #       const Locale &objectLocale,
        #       const Locale &displayLocale,
        #       UnicodeString &name
        # )
        result = BreakIterator.get_display_name(
            Locale.get_france(),
            us_locale,
            name)
        assert isinstance(result, UnicodeString)
        assert id(result) == id(name)
        assert name == "French (France)"

        name.remove()
        result = BreakIterator.get_display_name(
            "fr_FR",
            Locale("en_US"),
            name)
        assert isinstance(result, UnicodeString)
        assert id(result) == id(name)
        assert name == "French (France)"

        name.remove()
        result = BreakIterator.get_display_name(
            Locale("fr_FR"),
            "en_US",
            name)
        assert isinstance(result, UnicodeString)
        assert id(result) == id(name)
        assert name == "French (France)"

        name.remove()
        result = BreakIterator.get_display_name("fr_FR", "en_US", name)
        assert isinstance(result, UnicodeString)
        assert id(result) == id(name)
        assert name == "French (France)"

        # [2]
        # static UnicodeString &icu::BreakIterator::getDisplayName(
        #       const Locale &objectLocale,
        #       UnicodeString &name
        # )
        name.remove()
        result = BreakIterator.get_display_name(us_locale, name)
        assert isinstance(result, UnicodeString)
        assert id(result) == id(name)
        assert name == "English (United States)"

        name.remove()
        result = BreakIterator.get_display_name("en_US", name)
        assert isinstance(result, UnicodeString)
        assert id(result) == id(name)
        assert name == "English (United States)"
    finally:
        if default_locale:
            Locale.set_default(default_locale)


def test_get_locale():
    bi1 = BreakIterator.create_word_instance(Locale.get_english())
    bi2 = BreakIterator.create_word_instance(Locale.get_french())

    # Locale icu::BreakIterator::getLocale(
    #       ULocDataLocaleType type,
    #       UErrorCode &status
    # )
    loc = bi1.get_locale(ULocDataLocaleType.ULOC_VALID_LOCALE)
    assert isinstance(loc, Locale)
    assert loc.get_name() == "en"

    loc = bi1.get_locale(ULocDataLocaleType.ULOC_ACTUAL_LOCALE)
    assert isinstance(loc, Locale)
    assert len(loc.get_name()) == 0

    loc = bi2.get_locale(ULocDataLocaleType.ULOC_VALID_LOCALE)
    assert isinstance(loc, Locale)
    assert loc.get_name() == "fr"

    loc = bi2.get_locale(ULocDataLocaleType.ULOC_ACTUAL_LOCALE)
    assert isinstance(loc, Locale)
    assert len(loc.get_name()) == 0


@pytest.mark.skipif(U_ICU_VERSION_MAJOR_NUM < 52, reason="ICU4C<52")
def test_get_rule_status():
    bi = BreakIterator.create_word_instance(Locale.get_us())
    src = UnicodeString("foo bar baz!")
    bi.set_text(src)
    assert bi.first() == 0

    # int32_t icu::BreakIterator::getRuleStatus()
    assert bi.get_rule_status() == UWordBreak.UBRK_WORD_NONE
    assert bi.next() == 3
    assert bi.get_rule_status() == UWordBreak.UBRK_WORD_LETTER
    assert bi.next() == 4
    assert bi.get_rule_status() == UWordBreak.UBRK_WORD_NONE
    assert bi.next() == 7
    assert bi.get_rule_status() == UWordBreak.UBRK_WORD_LETTER
    assert bi.next() == 8
    assert bi.get_rule_status() == UWordBreak.UBRK_WORD_NONE
    assert bi.next() == 11
    assert bi.get_rule_status() == UWordBreak.UBRK_WORD_LETTER
    assert bi.next() == 12
    assert bi.get_rule_status() == UWordBreak.UBRK_WORD_NONE
    assert bi.next() == BreakIterator.DONE


@pytest.mark.skipif(U_ICU_VERSION_MAJOR_NUM < 52, reason="ICU4C<52")
def test_get_rule_status_vec():
    # From icu/source/test/intltest/rbbiapts.cpp:
    # void RBBIAPITest::TestRuleStatusVec()
    rules = UnicodeString("[A-N]{100}; \n"
                          "[a-w]{200}; \n"
                          "[\\p{L}]{300}; \n"
                          "[\\p{N}]{400}; \n"
                          "[0-5]{500}; \n"
                          "!.*;\n",
                          -1,
                          UnicodeString.INVARIANT)
    parse_error = UParseError()
    bi = RuleBasedBreakIterator(rules, parse_error)
    src = UnicodeString("Aapz5?")
    bi.set_text(src)

    # int32_t icu::BreakIterator::getRuleStatusVec(
    #       int32_t *fillInVec,
    #       int32_t capacity,
    #       UErrorCode &status
    # )
    assert bi.next() == 1
    values = bi.get_rule_status_vec()
    assert isinstance(values, list)
    assert values == [100, 300]

    assert bi.next() == 2
    values = bi.get_rule_status_vec()
    assert isinstance(values, list)
    assert values == [200, 300]

    assert bi.next() == 3
    values = bi.get_rule_status_vec()
    assert isinstance(values, list)
    assert values == [200, 300]

    assert bi.next() == 4
    values = bi.get_rule_status_vec()
    assert isinstance(values, list)
    assert values == [300]

    assert bi.next() == 5
    values = bi.get_rule_status_vec()
    assert isinstance(values, list)
    assert values == [400, 500]

    assert bi.next() == 6
    values = bi.get_rule_status_vec()
    assert isinstance(values, list)
    assert values == [0]


def test_get_rules():
    bi = BreakIterator.create_word_instance(Locale.get_us())

    # const UnicodeString &icu::RuleBasedBreakIterator::getRules(void)
    rules = bi.get_rules()
    assert isinstance(rules, UnicodeString)
    assert rules.length() > 0


def test_get_text():
    bi = BreakIterator.create_word_instance(Locale.get_us())

    # CharacterIterator &icu::BreakIterator::getText(void)
    # it = bi.get_text()  # Segmentation fault with Clang 12
    # assert it is None

    src = UnicodeString("foo bar baz.")
    bi.set_text(src)
    it = bi.get_text()
    assert isinstance(it, CharacterIterator)

    dest = UnicodeString()
    it.get_text(dest)
    assert dest == src


def test_get_utext():
    bi = BreakIterator.create_word_instance(Locale.get_us())

    # UText *icu::BreakIterator::getUText(
    #       UText *fillIn,
    #       UErrorCode &status
    # )
    ut1 = bi.get_utext(None)
    dest = utext_extract(ut1, 0, utext_native_length(ut1))
    assert len(dest) == 0
    utext_close(ut1)

    src1 = UnicodeString("ABC")
    bi.set_text(src1)
    ut1 = bi.get_utext(None)
    assert utext_extract(ut1, 0, utext_native_length(ut1)) == "ABC"

    src2 = UnicodeString("abc")
    bi.set_text(src2)
    ut2 = bi.get_utext(ut1)
    assert ut2 == ut1
    assert utext_extract(ut1, 0, utext_native_length(ut1)) == "abc"
    utext_close(ut1)


def test_hash_code():
    bi1 = BreakIterator.create_character_instance(Locale.get_us())
    bi2 = BreakIterator.create_word_instance(Locale.get_us())
    bi3 = BreakIterator.create_word_instance(Locale.get_us())

    # int32_t icu::RuleBasedBreakIterator::hashCode(void)
    assert bi1.hash_code() != bi2.hash_code()
    assert bi2.hash_code() == bi3.hash_code()

    src = UnicodeString("foo bar baz.")
    bi1.set_text(src)
    assert bi1.hash_code() != bi2.hash_code()
    assert bi2.hash_code() == bi3.hash_code()

    bi1.next()
    assert bi1.hash_code() != bi2.hash_code()
    assert bi2.hash_code() == bi3.hash_code()


def test_is_boundary():
    bi = BreakIterator.create_word_instance(Locale.get_us())
    src = UnicodeString("foo bar baz.")
    bi.set_text(src)

    # UBool icu::BreakIterator::isBoundary(int32_t offset)
    # [0, 3, 4, 7, 8, 11, 12]
    assert bi.is_boundary(0)
    assert not bi.is_boundary(1)
    assert not bi.is_boundary(2)
    assert bi.is_boundary(3)
    assert bi.is_boundary(4)
    assert not bi.is_boundary(5)
    assert not bi.is_boundary(6)
    assert bi.is_boundary(7)
    assert bi.is_boundary(8)
    assert not bi.is_boundary(9)
    assert not bi.is_boundary(10)
    assert bi.is_boundary(11)
    assert bi.is_boundary(12)


def test_next():
    bi = BreakIterator.create_word_instance(Locale.get_us())
    src = UnicodeString("foo bar baz.")
    bi.set_text(src)

    # [1]
    # int32_t icu::BreakIterator::next(int32_t n)
    # [0, 3, 4, 7, 8, 11, 12]
    assert bi.first() == 0
    assert bi.next(3) == 7
    assert bi.next(3) == 12
    assert bi.next(3) == BreakIterator.DONE
    assert bi.current() == 12
    assert bi.next(-3) == 7
    assert bi.next(-3) == 0
    assert bi.next(-3) == BreakIterator.DONE
    assert bi.current() == 0

    # [2]
    # int32_t icu::BreakIterator::next(void)
    assert bi.first() == 0
    assert bi.next() == 3
    assert bi.next() == 4
    assert bi.next() == 7
    assert bi.next() == 8
    assert bi.next() == 11
    assert bi.next() == 12
    assert bi.next() == BreakIterator.DONE


def test_operator():
    bi1 = BreakIterator.create_character_instance(Locale.get_us())
    bi2 = BreakIterator.create_word_instance(Locale.get_us())
    bi3 = BreakIterator.create_word_instance(Locale.get_us())
    src = UnicodeString("foo bar baz.")

    # UBool icu::BreakIterator::operator!=(const BreakIterator &rhs)
    assert bi1 != bi2
    assert bi1 != bi3

    # UBool icu::BreakIterator::operator==(const BreakIterator &)
    assert bi2 == bi3

    bi2.next()
    assert bi2 != bi3
    assert not (bi2 == bi3)

    bi3.next()
    assert not (bi2 != bi3)
    assert bi2 == bi3

    bi2.set_text(src)
    assert bi2 != bi3
    assert not (bi2 == bi3)

    bi3.set_text(src)
    assert not (bi2 != bi3)
    assert bi2 == bi3


def test_preceding():
    bi = BreakIterator.create_word_instance(Locale.get_us())
    src = UnicodeString("foo bar baz.")
    bi.set_text(src)

    # int32_t icu::BreakIterator::preceding(int32_t offset)
    # [0, 3, 4, 7, 8, 11, 12]
    assert bi.preceding(12) == 11
    assert bi.preceding(11) == 8
    assert bi.preceding(8) == 7
    assert bi.preceding(7) == 4
    assert bi.preceding(4) == 3
    assert bi.preceding(3) == 0
    assert bi.preceding(0) == BreakIterator.DONE


def test_previous():
    bi = BreakIterator.create_word_instance(Locale.get_us())
    src = UnicodeString("foo bar baz.")
    bi.set_text(src)

    # int32_t icu::BreakIterator::previous(void)
    assert bi.last() == 12
    assert bi.previous() == 11
    assert bi.previous() == 8
    assert bi.previous() == 7
    assert bi.previous() == 4
    assert bi.previous() == 3
    assert bi.previous() == 0
    assert bi.previous() == BreakIterator.DONE


def test_rule_based_break_iterator():
    src = UnicodeString("foo.")

    # [1]
    # icu::RuleBasedBreakIterator::RuleBasedBreakIterator()
    pass  # NotImplemented

    # [3]
    # icu::RuleBasedBreakIterator::RuleBasedBreakIterator(
    #       const UnicodeString &rules,
    #       UParseError &parseError,
    #       UErrorCode &status
    # )
    rules = UnicodeString("$dictionary = [a-z]; \n"
                          "!!forward; \n"
                          "$dictionary $dictionary; \n"
                          "!!reverse; \n"
                          "$dictionary $dictionary; \n")
    parse_error = UParseError()
    bi3 = RuleBasedBreakIterator(rules, parse_error)
    bi3.set_text(src)
    assert bi3.first() == 0
    assert bi3.next() == 2
    assert bi3.next() == 3
    assert bi3.next() == 4
    assert bi3.next() == BreakIterator.DONE
    assert bi3.current() == 4

    bi3a = RuleBasedBreakIterator(
        "$dictionary = [a-z]; \n"
        "!!forward; \n"
        "$dictionary $dictionary; \n"
        "!!reverse; \n"
        "$dictionary $dictionary; \n",
        parse_error)
    bi3a.set_text(src)
    assert bi3a.first() == 0
    assert bi3a.next() == 2
    assert bi3a.next() == 3
    assert bi3a.next() == 4
    assert bi3a.next() == BreakIterator.DONE
    assert bi3a.current() == 4
    assert bi3 == bi3a

    # [4]
    # icu::RuleBasedBreakIterator::RuleBasedBreakIterator(
    #       const uint8_t *compiledRules,
    #       uint32_t ruleLength,
    #       UErrorCode &status
    # )
    binary_rules = bi3.get_binary_rules()
    assert isinstance(binary_rules, list)
    assert len(binary_rules) > 0
    bi4 = RuleBasedBreakIterator(binary_rules, len(binary_rules))
    bi4.set_text(src)
    assert bi4.first() == 0
    assert bi4.next() == 2
    assert bi4.next() == 3
    assert bi4.next() == 4
    assert bi4.next() == BreakIterator.DONE
    assert bi4.current() == 4

    # [2]
    # icu::RuleBasedBreakIterator::RuleBasedBreakIterator(
    #       const RuleBasedBreakIterator &that
    # )
    bi2 = RuleBasedBreakIterator(bi3)
    assert bi3.current() == 4
    assert bi2.current() == 4
    assert bi2.first() == 0
    assert bi2.next() == 2
    assert bi2.next() == 3
    assert bi2.next() == 4
    assert bi2.next() == BreakIterator.DONE
    assert bi2.current() == 4

    # [5]
    # icu::RuleBasedBreakIterator::RuleBasedBreakIterator(
    #       UDataMemory *image,
    #       UErrorCode &status
    # )
    pass  # NotImplemented


def test_set_text():
    bi = BreakIterator.create_word_instance(Locale.get_us())
    src1 = UnicodeString("foo bar baz.")
    src2 = UnicodeString("lorem ipsum")

    # [1]
    # void icu::BreakIterator::setText(const UnicodeString &newText)
    bi.set_text(src1)
    assert bi.current() == 0
    assert bi.next() == 3

    bi.set_text(src2)
    assert bi.current() == 0
    assert bi.next() == 5

    # [2]
    # void icu::BreakIterator::setText(
    #       UText *text,
    #       UErrorCode &status
    # )
    ut = utext_open_const_unicode_string(None, src1)
    bi.set_text(ut)
    assert bi.current() == 0
    assert bi.next() == 3

    utext_open_const_unicode_string(ut, src2)
    bi.set_text(ut)
    assert bi.current() == 0
    assert bi.next() == 5
    utext_close(ut)



import pytest

from icupy import (
    ICUError, UErrorCode, UnicodeString, UScriptCode, UScriptUsage,
    uscript_breaks_between_letters, uscript_get_code, uscript_get_name,
    uscript_get_sample_string, uscript_get_sample_unicode_string,
    uscript_get_script, uscript_get_script_extensions, uscript_get_short_name,
    uscript_get_usage, uscript_has_script, uscript_is_cased,
    uscript_is_right_to_left,
)


def test_api():
    assert uscript_breaks_between_letters(UScriptCode.USCRIPT_JAPANESE)
    assert not uscript_breaks_between_letters(UScriptCode.USCRIPT_LATIN)

    # int32_t uscript_getCode(const char *nameOrAbbrOrLocale,
    #                         UScriptCode *fillIn,
    #                         int32_t capacity,
    #                         UErrorCode *err
    # )
    fill_in = uscript_get_code("ja_JP")
    assert isinstance(fill_in, list)
    assert len(fill_in) == 3
    assert UScriptCode.USCRIPT_HAN in fill_in
    assert UScriptCode.USCRIPT_HIRAGANA in fill_in
    assert UScriptCode.USCRIPT_KATAKANA in fill_in

    assert uscript_get_name(UScriptCode.USCRIPT_JAPANESE) == "Jpan"
    assert uscript_get_name(UScriptCode.USCRIPT_MALAYALAM) == "Malayalam"
    assert uscript_get_name(UScriptCode.USCRIPT_INVALID_CODE) is None

    # int32_t uscript_getSampleString(UScriptCode script,
    #                                 UChar *dest,
    #                                 int32_t capacity,
    #                                 UErrorCode *pErrorCode
    # )
    dest = uscript_get_sample_string(UScriptCode.USCRIPT_JAPANESE)
    assert isinstance(dest, str)
    assert len(dest) > 0

    result = uscript_get_sample_unicode_string(UScriptCode.USCRIPT_JAPANESE)
    assert isinstance(result, UnicodeString)
    assert dest == result

    # UScriptCode uscript_getScript(UChar32 codepoint, UErrorCode *err)
    assert uscript_get_script(0x0d02) == UScriptCode.USCRIPT_MALAYALAM
    assert uscript_get_script(0) == UScriptCode.USCRIPT_COMMON
    with pytest.raises(ICUError) as exc_info:
        _ = uscript_get_script(-1)
    assert exc_info.value.args[0] == UErrorCode.U_ILLEGAL_ARGUMENT_ERROR

    # int32_t uscript_getScriptExtensions(UChar32 c,
    #                                     UScriptCode *scripts,
    #                                     int32_t capacity,
    #                                     UErrorCode *errorCode
    # )
    scripts = uscript_get_script_extensions(0x3031)
    assert isinstance(scripts, list)
    assert len(scripts) == 2
    assert UScriptCode.USCRIPT_HIRAGANA in scripts
    assert UScriptCode.USCRIPT_KATAKANA in scripts

    scripts = uscript_get_script_extensions(-1)
    assert isinstance(scripts, list)
    assert len(scripts) == 1
    assert UScriptCode.USCRIPT_UNKNOWN in scripts

    assert uscript_get_short_name(UScriptCode.USCRIPT_JAPANESE) == "Jpan"
    assert uscript_get_short_name(UScriptCode.USCRIPT_MALAYALAM) == "Mlym"
    assert uscript_get_short_name(UScriptCode.USCRIPT_INVALID_CODE) is None

    assert uscript_get_usage(
        UScriptCode.USCRIPT_HIRAGANA
    ) == UScriptUsage.USCRIPT_USAGE_RECOMMENDED
    assert uscript_get_usage(
        UScriptCode.USCRIPT_ADLAM
    ) == UScriptUsage.USCRIPT_USAGE_LIMITED_USE

    assert uscript_has_script(0x3031, UScriptCode.USCRIPT_HIRAGANA)
    assert not uscript_has_script(0x3031, UScriptCode.USCRIPT_HAN)

    assert uscript_is_cased(UScriptCode.USCRIPT_LATIN)
    assert not uscript_is_cased(UScriptCode.USCRIPT_HIRAGANA)

    assert uscript_is_right_to_left(UScriptCode.USCRIPT_ARABIC)
    assert not uscript_is_right_to_left(UScriptCode.USCRIPT_LATIN)



import pytest
from icupy import U_ICU_VERSION_MAJOR_NUM

if U_ICU_VERSION_MAJOR_NUM < 53:
    pytest.skip("ICU4C<53", allow_module_level=True)

from icupy import (
    DecimalFormat, Locale, NumberFormat, RelativeDateTimeFormatter,
    UDateAbsoluteUnit, UDateDirection, UDateRelativeUnit, UnicodeString,
)


def test_api():
    fmt = RelativeDateTimeFormatter("en")

    # UnicodeString &icu::RelativeDateTimeFormatter::combineDateAndTime(
    #       const UnicodeString &relativeDateString,
    #       const UnicodeString &timeString,
    #       UnicodeString &appendTo,
    #       UErrorCode &status
    # )
    append_to = UnicodeString()
    result = fmt.combine_date_and_time(
        UnicodeString("yesterday"),
        UnicodeString("3:50"),
        append_to)
    assert isinstance(result, UnicodeString)
    assert id(result) == id(append_to)
    assert result == "yesterday, 3:50"

    append_to.remove()
    result = fmt.combine_date_and_time(
        "yesterday",
        UnicodeString("3:50"),
        append_to)
    assert isinstance(result, UnicodeString)
    assert id(result) == id(append_to)
    assert result == "yesterday, 3:50"

    append_to.remove()
    result = fmt.combine_date_and_time(
        UnicodeString("yesterday"),
        "3:50",
        append_to)
    assert isinstance(result, UnicodeString)
    assert id(result) == id(append_to)
    assert result == "yesterday, 3:50"

    append_to.remove()
    result = fmt.combine_date_and_time(
        "yesterday",
        "3:50",
        append_to)
    assert isinstance(result, UnicodeString)
    assert id(result) == id(append_to)
    assert result == "yesterday, 3:50"

    # const NumberFormat &icu::RelativeDateTimeFormatter::getNumberFormat()
    numfmt = fmt.get_number_format()
    assert isinstance(numfmt, NumberFormat)


@pytest.mark.skipif(U_ICU_VERSION_MAJOR_NUM < 54, reason="ICU4C<54")
def test_api_54():
    from icupy import UDateRelativeDateTimeFormatterStyle, UDisplayContext

    fmt = RelativeDateTimeFormatter(
        "en",
        None,
        UDateRelativeDateTimeFormatterStyle.UDAT_STYLE_NARROW,
        UDisplayContext.UDISPCTX_CAPITALIZATION_FOR_MIDDLE_OF_SENTENCE)

    # UDisplayContext
    # icu::RelativeDateTimeFormatter::getCapitalizationContext()
    result = fmt.get_capitalization_context()
    assert isinstance(result, UDisplayContext)
    assert (result
            == UDisplayContext.UDISPCTX_CAPITALIZATION_FOR_MIDDLE_OF_SENTENCE)

    # UDateRelativeDateTimeFormatterStyle
    # icu::RelativeDateTimeFormatter::getFormatStyle()
    result = fmt.get_format_style()
    assert isinstance(result, UDateRelativeDateTimeFormatterStyle)
    assert result == UDateRelativeDateTimeFormatterStyle.UDAT_STYLE_NARROW


def test_format():
    fmt = RelativeDateTimeFormatter("en")

    # [2]
    # UnicodeString &icu::RelativeDateTimeFormatter::format(
    #       double quantity,
    #       UDateDirection direction,
    #       UDateRelativeUnit unit,
    #       UnicodeString &appendTo,
    #       UErrorCode &status
    # )
    append_to = UnicodeString()
    result = fmt.format(
        0.5,
        UDateDirection.UDAT_DIRECTION_LAST,
        UDateRelativeUnit.UDAT_RELATIVE_SECONDS,
        append_to)
    assert isinstance(result, UnicodeString)
    assert id(result) == id(append_to)
    assert result == "0.5 seconds ago"

    append_to.remove()
    result = fmt.format(
        1,
        UDateDirection.UDAT_DIRECTION_LAST,
        UDateRelativeUnit.UDAT_RELATIVE_SECONDS,
        append_to)
    assert isinstance(result, UnicodeString)
    assert id(result) == id(append_to)
    assert result == "1 second ago"

    # [3]
    # UnicodeString &icu::RelativeDateTimeFormatter::format(
    #       UDateDirection direction,
    #       UDateAbsoluteUnit unit,
    #       UnicodeString &appendTo,
    #       UErrorCode &status
    # )
    append_to.remove()
    result = fmt.format(
        UDateDirection.UDAT_DIRECTION_NEXT,
        UDateAbsoluteUnit.UDAT_ABSOLUTE_DAY,
        append_to)
    assert isinstance(result, UnicodeString)
    assert id(result) == id(append_to)
    assert result == "tomorrow"


@pytest.mark.skipif(U_ICU_VERSION_MAJOR_NUM < 57, reason="ICU4C<57")
def test_format_57():
    from icupy import URelativeDateTimeUnit

    fmt = RelativeDateTimeFormatter("en")

    # [1]
    # UnicodeString &icu::RelativeDateTimeFormatter::format(
    #       double offset,
    #       URelativeDateTimeUnit unit,
    #       UnicodeString &appendTo,
    #       UErrorCode &status
    # )
    append_to = UnicodeString()
    result = fmt.format(
        0.5,
        URelativeDateTimeUnit.UDAT_REL_UNIT_SECOND,
        append_to)
    assert isinstance(result, UnicodeString)
    assert id(result) == id(append_to)
    assert result == "in 0.5 seconds"

    append_to.remove()
    result = fmt.format(
        1,
        URelativeDateTimeUnit.UDAT_REL_UNIT_SECOND,
        append_to)
    assert isinstance(result, UnicodeString)
    assert id(result) == id(append_to)
    assert result == "in 1 second"


@pytest.mark.skipif(U_ICU_VERSION_MAJOR_NUM < 57, reason="ICU4C<57")
def test_format_numeric():
    from icupy import URelativeDateTimeUnit

    fmt = RelativeDateTimeFormatter("en-US")

    # UnicodeString &icu::RelativeDateTimeFormatter::formatNumeric(
    #       double offset,
    #       URelativeDateTimeUnit unit,
    #       UnicodeString &appendTo,
    #       UErrorCode &status
    # )
    append_to = UnicodeString()
    result = fmt.format_numeric(
        0.5,
        URelativeDateTimeUnit.UDAT_REL_UNIT_SECOND,
        append_to)
    assert isinstance(result, UnicodeString)
    assert id(result) == id(append_to)
    assert result == "in 0.5 seconds"

    append_to.remove()
    result = fmt.format_numeric(
        1,
        URelativeDateTimeUnit.UDAT_REL_UNIT_SECOND,
        append_to)
    assert isinstance(result, UnicodeString)
    assert id(result) == id(append_to)
    assert result == "in 1 second"


@pytest.mark.skipif(U_ICU_VERSION_MAJOR_NUM < 64, reason="ICU4C<64")
def test_format_numeric_to_value():
    from icupy import (
        ConstrainedFieldPosition, FormattedRelativeDateTime,
        FormattedValue, UFieldCategory, UNumberFormatFields,
        URelativeDateTimeFormatterField, URelativeDateTimeUnit,
        UnicodeStringAppendable,
    )

    assert issubclass(FormattedRelativeDateTime, FormattedValue)
    fmt = RelativeDateTimeFormatter("en-US")

    # FormattedRelativeDateTime
    # icu::RelativeDateTimeFormatter::formatNumericToValue(
    #       double offset,
    #       URelativeDateTimeUnit unit,
    #       UErrorCode &status
    # )
    fv = fmt.format_numeric_to_value(
        1.5,
        URelativeDateTimeUnit.UDAT_REL_UNIT_WEEK)
    assert isinstance(fv, FormattedRelativeDateTime)

    # Appendable &icu::FormattedRelativeDateTime::appendTo(
    #       Appendable &appendable,
    #       UErrorCode &status
    # )
    s = UnicodeString("x")
    appendable = UnicodeStringAppendable(s)
    result = fv.append_to(appendable)
    assert isinstance(result, UnicodeStringAppendable)
    assert id(result) == id(appendable)
    assert s == "xin 1.5 weeks"

    # UBool icu::FormattedRelativeDateTime::nextPosition(
    #       ConstrainedFieldPosition &cfpos,
    #       UErrorCode &status
    # )
    cfpos = ConstrainedFieldPosition()
    assert fv.next_position(cfpos)
    assert (cfpos.get_category()
            == UFieldCategory.UFIELD_CATEGORY_RELATIVE_DATETIME)
    assert (cfpos.get_field()
            == URelativeDateTimeFormatterField.UDAT_REL_LITERAL_FIELD)
    assert (cfpos.get_start(), cfpos.get_limit()) == (0, 2)

    assert fv.next_position(cfpos)
    assert cfpos.get_category() == UFieldCategory.UFIELD_CATEGORY_NUMBER
    assert cfpos.get_field() == UNumberFormatFields.UNUM_INTEGER_FIELD
    assert (cfpos.get_start(), cfpos.get_limit()) == (3, 4)

    assert fv.next_position(cfpos)
    assert cfpos.get_category() == UFieldCategory.UFIELD_CATEGORY_NUMBER
    assert (cfpos.get_field()
            == UNumberFormatFields.UNUM_DECIMAL_SEPARATOR_FIELD)
    assert (cfpos.get_start(), cfpos.get_limit()) == (4, 5)

    assert fv.next_position(cfpos)
    assert cfpos.get_category() == UFieldCategory.UFIELD_CATEGORY_NUMBER
    assert cfpos.get_field() == UNumberFormatFields.UNUM_FRACTION_FIELD
    assert (cfpos.get_start(), cfpos.get_limit()) == (5, 6)

    assert fv.next_position(cfpos)
    assert (cfpos.get_category()
            == UFieldCategory.UFIELD_CATEGORY_RELATIVE_DATETIME)
    assert (cfpos.get_field()
            == URelativeDateTimeFormatterField.UDAT_REL_NUMERIC_FIELD)
    assert (cfpos.get_start(), cfpos.get_limit()) == (3, 6)

    assert fv.next_position(cfpos)
    assert (cfpos.get_category()
            == UFieldCategory.UFIELD_CATEGORY_RELATIVE_DATETIME)
    assert (cfpos.get_field()
            == URelativeDateTimeFormatterField.UDAT_REL_LITERAL_FIELD)
    assert (cfpos.get_start(), cfpos.get_limit()) == (7, 12)

    assert not fv.next_position(cfpos)

    # UnicodeString icu::FormattedRelativeDateTime::toString(
    #       UErrorCode &status
    # )
    result = fv.to_string()
    assert isinstance(result, UnicodeString)
    assert result == "in 1.5 weeks"

    # UnicodeString icu::FormattedRelativeDateTime::toTempString(
    #       UErrorCode &status
    # )
    result = fv.to_temp_string()
    assert isinstance(result, UnicodeString)
    assert result == "in 1.5 weeks"

    fv = fmt.format_numeric_to_value(
        1,
        URelativeDateTimeUnit.UDAT_REL_UNIT_WEEK)
    assert isinstance(fv, FormattedRelativeDateTime)
    assert fv.to_temp_string() == "in 1 week"


@pytest.mark.skipif(U_ICU_VERSION_MAJOR_NUM < 64, reason="ICU4C<64")
def test_format_to_value():
    from icupy import FormattedRelativeDateTime, URelativeDateTimeUnit

    fmt = RelativeDateTimeFormatter("en-US")

    # [1]
    # FormattedRelativeDateTime icu::RelativeDateTimeFormatter::formatToValue(
    #       double offset,
    #       URelativeDateTimeUnit unit,
    #       UErrorCode &status
    # )
    fv = fmt.format_to_value(1.0, URelativeDateTimeUnit.UDAT_REL_UNIT_DAY)
    assert isinstance(fv, FormattedRelativeDateTime)
    assert fv.to_temp_string() == "tomorrow"

    fv = fmt.format_to_value(3, URelativeDateTimeUnit.UDAT_REL_UNIT_DAY)
    assert isinstance(fv, FormattedRelativeDateTime)
    assert fv.to_temp_string() == "in 3 days"

    # [2]
    # FormattedRelativeDateTime icu::RelativeDateTimeFormatter::formatToValue(
    #       double quantity,
    #       UDateDirection direction,
    #       UDateRelativeUnit unit,
    #       UErrorCode &status
    # )
    fv = fmt.format_to_value(
        12.0,
        UDateDirection.UDAT_DIRECTION_LAST,
        UDateRelativeUnit.UDAT_RELATIVE_HOURS)
    assert isinstance(fv, FormattedRelativeDateTime)
    assert fv.to_temp_string() == "12 hours ago"

    fv = fmt.format_to_value(
        12,
        UDateDirection.UDAT_DIRECTION_LAST,
        UDateRelativeUnit.UDAT_RELATIVE_HOURS)
    assert isinstance(fv, FormattedRelativeDateTime)
    assert fv.to_temp_string() == "12 hours ago"

    # [3]
    # FormattedRelativeDateTime icu::RelativeDateTimeFormatter::formatToValue(
    #       UDateDirection direction,
    #       UDateAbsoluteUnit unit,
    #       UErrorCode &status
    # )
    fv = fmt.format_to_value(
        UDateDirection.UDAT_DIRECTION_NEXT,
        UDateAbsoluteUnit.UDAT_ABSOLUTE_MONDAY)
    assert isinstance(fv, FormattedRelativeDateTime)
    assert fv.to_temp_string() == "next Monday"


def test_relative_date_time_formatter():
    # [1]
    # icu::RelativeDateTimeFormatter::RelativeDateTimeFormatter(
    #       UErrorCode &status
    # )
    fmt1 = RelativeDateTimeFormatter()

    # [2]
    # icu::RelativeDateTimeFormatter::RelativeDateTimeFormatter(
    #       const Locale &locale,
    #       UErrorCode &status
    # )
    _ = RelativeDateTimeFormatter(Locale("en"))
    _ = RelativeDateTimeFormatter("en")

    # [3]
    # icu::RelativeDateTimeFormatter::RelativeDateTimeFormatter(
    #       const Locale &locale,
    #       NumberFormat *nfToAdopt,
    #       UErrorCode &status
    # )
    _ = RelativeDateTimeFormatter(Locale("en"), DecimalFormat("0"))
    _ = RelativeDateTimeFormatter(Locale("en"), None)
    _ = RelativeDateTimeFormatter("en", DecimalFormat("0"))
    _ = RelativeDateTimeFormatter("en", None)

    # [5]
    # icu::RelativeDateTimeFormatter::RelativeDateTimeFormatter(
    #       const RelativeDateTimeFormatter &other
    # )
    _ = RelativeDateTimeFormatter(fmt1)


@pytest.mark.skipif(U_ICU_VERSION_MAJOR_NUM < 54, reason="ICU4C<54")
def test_relative_date_time_formatter_54():
    from icupy import UDateRelativeDateTimeFormatterStyle, UDisplayContext

    # [4]
    # icu::RelativeDateTimeFormatter::RelativeDateTimeFormatter(
    #       const Locale &locale,
    #       NumberFormat *nfToAdopt,
    #       UDateRelativeDateTimeFormatterStyle style,
    #       UDisplayContext capitalizationContext,
    #       UErrorCode &status
    # )
    _ = RelativeDateTimeFormatter(
        Locale("en"),
        DecimalFormat("0"),
        UDateRelativeDateTimeFormatterStyle.UDAT_STYLE_LONG,
        UDisplayContext.UDISPCTX_CAPITALIZATION_FOR_BEGINNING_OF_SENTENCE)

    _ = RelativeDateTimeFormatter(
        Locale("en"),
        None,
        UDateRelativeDateTimeFormatterStyle.UDAT_STYLE_LONG,
        UDisplayContext.UDISPCTX_CAPITALIZATION_FOR_BEGINNING_OF_SENTENCE)

    _ = RelativeDateTimeFormatter(
        "en",
        DecimalFormat("0"),
        UDateRelativeDateTimeFormatterStyle.UDAT_STYLE_LONG,
        UDisplayContext.UDISPCTX_CAPITALIZATION_FOR_BEGINNING_OF_SENTENCE)

    _ = RelativeDateTimeFormatter(
        "en",
        None,
        UDateRelativeDateTimeFormatterStyle.UDAT_STYLE_LONG,
        UDisplayContext.UDISPCTX_CAPITALIZATION_FOR_BEGINNING_OF_SENTENCE)



import pytest
from icupy import (
    ConstVoidPtr, ICUError, RegexMatcher, RegexPattern, UErrorCode,
    URegexFindProgressCallbackPtr, URegexMatchCallbackPtr, URegexpFlag,
    UTextVector, UnicodeString, UnicodeStringVector,
    utext_close, utext_extract, utext_get_native_index, utext_native_length,
    utext_open_const_unicode_string, utext_open_unicode_string,
    utext_open_utf8, utext_replace,
)

from . import gc


def test_api():
    regexp = UnicodeString("\\w+")
    src = UnicodeString("foo bar baz")
    matcher = RegexMatcher(regexp, src, 0)

    s = UnicodeString()
    with gc(utext_open_unicode_string(None, s), utext_close) as dest:
        # UText *icu::RegexMatcher::getInput(
        #       UText *dest,
        #       UErrorCode &status
        # )
        result = matcher.get_input(dest)
        assert result == dest
        assert utext_extract(dest, 0, utext_native_length(dest)) == src

        result = matcher.get_input(None)
        assert utext_extract(result, 0, utext_native_length(result)) == src
        utext_close(result)

    # int32_t icu::RegexMatcher::getStackLimit()
    assert matcher.get_stack_limit() > 0

    # void icu::RegexMatcher::setStackLimit(
    #       int32_t limit,
    #       UErrorCode &status
    # )
    matcher.set_stack_limit(0)
    assert matcher.get_stack_limit() == 0

    # int32_t icu::RegexMatcher::getTimeLimit()
    assert matcher.get_time_limit() == 0

    # void icu::RegexMatcher::setTimeLimit(
    #       int32_t limit,
    #       UErrorCode &status
    # )
    matcher.set_time_limit(10)
    assert matcher.get_time_limit() == 10

    # UBool icu::RegexMatcher::hasAnchoringBounds()
    assert matcher.has_anchoring_bounds()

    # RegexMatcher &icu::RegexMatcher::useAnchoringBounds(UBool b)
    result = matcher.use_anchoring_bounds(False)
    assert isinstance(result, RegexMatcher)
    assert result == matcher
    assert not matcher.has_anchoring_bounds()

    # UBool icu::RegexMatcher::hasTransparentBounds()
    assert not matcher.has_transparent_bounds()

    # RegexMatcher &icu::RegexMatcher::useTransparentBounds(UBool b)
    result = matcher.use_transparent_bounds(True)
    assert isinstance(result, RegexMatcher)
    assert result == matcher
    assert matcher.has_transparent_bounds()

    matcher.use_anchoring_bounds(True)
    matcher.use_transparent_bounds(False)
    assert matcher.find()  # "foo"

    # UBool icu::RegexMatcher::hitEnd()
    assert not matcher.hit_end()

    # UBool icu::RegexMatcher::requireEnd()
    assert not matcher.require_end()

    assert matcher.find()  # "bar"
    assert not matcher.hit_end()
    assert not matcher.require_end()

    assert matcher.find()  # "baz"
    assert matcher.hit_end()
    assert not matcher.require_end()

    assert not matcher.find()
    assert matcher.hit_end()
    assert not matcher.require_end()

    # const UnicodeString &icu::RegexMatcher::input()
    result = matcher.input()
    assert isinstance(result, UnicodeString)
    assert result == src

    # UText *icu::RegexMatcher::inputText()
    result = matcher.input_text()
    assert utext_extract(result, 0, utext_native_length(result)) == src

    # const RegexPattern &icu::RegexMatcher::pattern()
    pattern = matcher.pattern()
    assert isinstance(pattern, RegexPattern)
    assert pattern.pattern() == regexp


def test_append_replacement():
    regexp = UnicodeString("\\w+")
    src = UnicodeString("foo bar baz")
    matcher = RegexMatcher(regexp, src, 0)

    assert matcher.find()
    assert matcher.find()

    # [1]
    # RegexMatcher &icu::RegexMatcher::appendReplacement(
    #       UnicodeString &dest,
    #       const UnicodeString &replacement,
    #       UErrorCode &status
    # )
    dest1 = UnicodeString()
    replacement1 = UnicodeString("abc")
    result = matcher.append_replacement(dest1, replacement1)
    assert isinstance(result, RegexMatcher)
    assert id(result) == id(matcher)
    assert dest1 == "foo abc"

    matcher.reset()
    assert matcher.find()
    assert matcher.find()

    dest1.remove()
    result = matcher.append_replacement(dest1, "abc")
    assert isinstance(result, RegexMatcher)
    assert id(result) == id(matcher)
    assert dest1 == "foo abc"

    matcher.reset()
    assert matcher.find()
    assert matcher.find()

    # [2]
    # RegexMatcher &icu::RegexMatcher::appendReplacement(
    #       UText *dest,
    #       UText *replacement,
    #       UErrorCode &status
    # )
    s2 = UnicodeString()
    dest2 = utext_open_unicode_string(None, s2)
    replacement2 = utext_open_utf8(None, "abc", -1)
    result = matcher.append_replacement(dest2, replacement2)
    assert isinstance(result, RegexMatcher)
    assert id(result) == id(matcher)
    assert utext_extract(dest2, 0, utext_native_length(dest2)) == "foo abc"
    utext_close(dest2)
    utext_close(replacement2)


def test_append_tail():
    regexp = UnicodeString("\\w+")
    src = UnicodeString("foo bar baz")
    matcher = RegexMatcher(regexp, src, 0)

    assert matcher.find()
    assert matcher.find()

    # [1]
    # UnicodeString &icu::RegexMatcher::appendTail(UnicodeString &dest)
    dest1 = UnicodeString()
    replacement1 = UnicodeString("abc")
    matcher.append_replacement(dest1, replacement1)
    result = matcher.append_tail(dest1)
    assert isinstance(result, UnicodeString)
    assert id(result) == id(dest1)
    assert dest1 == "foo abc baz"

    matcher.reset()
    assert matcher.find()
    assert matcher.find()

    # [2]
    # UText *icu::RegexMatcher::appendTail(
    #       UText *dest,
    #       UErrorCode &status
    # )
    s2 = UnicodeString()
    dest2 = utext_open_unicode_string(None, s2)
    replacement2 = utext_open_utf8(None, "abc", -1)
    matcher.append_replacement(dest2, replacement2)
    result = matcher.append_tail(dest2)
    assert result == dest2
    assert utext_extract(dest2, 0, utext_native_length(dest2)) == "foo abc baz"
    utext_close(dest2)
    utext_close(replacement2)


def test_end():
    regexp = UnicodeString("01(23(45)67)(.*)")
    src = UnicodeString("0123456789")
    matcher = RegexMatcher(regexp, src, 0)

    matcher.looking_at(0)

    # [1]
    # int32_t icu::RegexMatcher::end(
    #       int32_t group,
    #       UErrorCode &status
    # )
    assert matcher.end(0) == 10
    assert matcher.end(1) == 8
    assert matcher.end(2) == 6
    assert matcher.end(3) == 10

    # int64_t icu::RegexMatcher::end64(
    #       int32_t group,
    #       UErrorCode &status
    # )
    assert matcher.end64(0) == 10
    assert matcher.end64(1) == 8
    assert matcher.end64(2) == 6
    assert matcher.end64(3) == 10

    # [2]
    # int32_t icu::RegexMatcher::end(UErrorCode &status)
    assert matcher.end() == 10

    # int64_t icu::RegexMatcher::end64(UErrorCode &status)
    assert matcher.end64() == 10


def test_find():
    regexp = UnicodeString("abc")
    src = UnicodeString(".abc..abc..")
    matcher = RegexMatcher(regexp, src, 0)

    # [1]
    # UBool icu::RegexMatcher::find()
    #
    # [3]
    # UBool icu::RegexMatcher::find(UErrorCode &status)
    # (ICU>=55)
    assert matcher.find()
    assert matcher.start() == 1
    assert matcher.find()
    assert matcher.start() == 6
    assert not matcher.find()

    # [2]
    # UBool icu::RegexMatcher::find(
    #       int64_t start,
    #       UErrorCode &status
    # )
    assert matcher.find(2)
    assert matcher.start() == 6
    assert matcher.find(6)
    assert matcher.start() == 6
    assert not matcher.find(7)


def test_group():
    regexp = UnicodeString("01(23(45)67)(.*)")
    src = UnicodeString("0123456789")
    matcher = RegexMatcher(regexp, src, 0)

    # int32_t icu::RegexMatcher::groupCount()
    assert matcher.group_count() == 3
    matcher.looking_at(0)

    # [1]
    # UnicodeString icu::RegexMatcher::group(
    #       int32_t groupNum,
    #       UErrorCode &status
    # )
    result = matcher.group(0)
    assert isinstance(result, UnicodeString)
    assert result == "0123456789"
    assert matcher.group(1) == "234567"
    assert matcher.group(2) == "45"
    assert matcher.group(3) == "89"

    # [2]
    # UText *icu::RegexMatcher::group(
    #       int32_t groupNum,
    #       UText *dest,
    #       int64_t &group_len,
    #       UErrorCode &status
    # )
    dest, group_len = matcher.group(0, None)
    assert utext_get_native_index(dest) == 0
    assert group_len == 10
    assert utext_extract(dest, 0, 10) == "0123456789"

    result, group_len = matcher.group(1, dest)
    assert result == dest
    assert utext_get_native_index(dest) == 2
    assert group_len == 6
    assert utext_extract(dest, 2, 2 + 6) == "234567"

    result, group_len = matcher.group(2, dest)
    assert utext_get_native_index(dest) == 4
    assert group_len == 2
    assert utext_extract(dest, 4, 4 + 2) == "45"

    result, group_len = matcher.group(3, dest)
    assert utext_get_native_index(dest) == 8
    assert group_len == 2
    assert utext_extract(dest, 8, 8 + 2) == "89"
    utext_close(dest)

    # [3]
    # UnicodeString icu::RegexMatcher::group(UErrorCode &status)
    result = matcher.group()
    assert isinstance(result, UnicodeString)
    assert result == "0123456789"

    # [4]
    # UText *icu::RegexMatcher::group(
    #       UText *dest,
    #       int64_t &group_len,
    #       UErrorCode &status
    # )
    dest, group_len = matcher.group(None)
    assert utext_get_native_index(dest) == 0
    assert group_len == 10
    assert utext_extract(dest, 0, 10) == "0123456789"

    result, group_len = matcher.group(dest)
    assert result == dest
    assert utext_get_native_index(dest) == 0
    assert group_len == 10
    assert utext_extract(dest, 0, 10) == "0123456789"
    utext_close(dest)


def test_looking_at():
    regexp = UnicodeString("\\w+")
    src = UnicodeString("foo bar baz")
    matcher = RegexMatcher(regexp, src, 0)

    # [1]
    # UBool icu::RegexMatcher::lookingAt(
    #       int64_t startIndex,
    #       UErrorCode &status
    # )
    assert matcher.looking_at(0)
    assert matcher.looking_at(1)
    assert matcher.looking_at(2)
    assert not matcher.looking_at(3)

    # [2]
    # UBool icu::RegexMatcher::lookingAt(UErrorCode &status)
    assert matcher.looking_at()


def test_matches():
    regexp = UnicodeString("\\w+")
    src = UnicodeString("foo bar baz")
    matcher = RegexMatcher(regexp, src, 0)

    # [1]
    # UBool icu::RegexMatcher::matches(
    #       int64_t startIndex,
    #       UErrorCode &status
    # )
    assert not matcher.matches(0)
    assert not matcher.matches(4)
    assert matcher.matches(8)

    # [2]
    # UBool icu::RegexMatcher::matches(UErrorCode &status)
    assert not matcher.matches()


def test_regex_matcher():
    s = "\\w+"

    # [1]
    # icu::RegexMatcher::RegexMatcher(
    #       const UnicodeString &regexp,
    #       uint32_t flags,
    #       UErrorCode &status
    # )
    regexp1 = UnicodeString(s)
    test1 = RegexMatcher(regexp1, URegexpFlag.UREGEX_CASE_INSENSITIVE)
    pattern1 = test1.pattern()
    assert pattern1.pattern() == s

    test1a = RegexMatcher(s, URegexpFlag.UREGEX_CASE_INSENSITIVE)
    assert test1a.pattern().pattern() == s

    # [2]
    # icu::RegexMatcher::RegexMatcher(
    #       UText *regexp,
    #       uint32_t flags,
    #       UErrorCode &status
    # )
    regexp2 = utext_open_utf8(None, s, -1)
    test2 = RegexMatcher(regexp2, URegexpFlag.UREGEX_CASE_INSENSITIVE)
    pattern2 = test2.pattern()
    assert pattern2.pattern() == s
    utext_close(regexp2)

    # [3]
    # icu::RegexMatcher::RegexMatcher(
    #       const UnicodeString &regexp,
    #       const UnicodeString &input,
    #       uint32_t flags,
    #       UErrorCode &status
    # )
    regexp3 = UnicodeString(s)
    src3 = UnicodeString("foo bar baz")
    test3 = RegexMatcher(regexp3, src3, URegexpFlag.UREGEX_CASE_INSENSITIVE)
    pattern3 = test3.pattern()
    assert pattern3.pattern() == s
    assert test3.input() == "foo bar baz"

    test3a = RegexMatcher(
        s,
        src3,
        URegexpFlag.UREGEX_CASE_INSENSITIVE)
    assert test3a.pattern().pattern() == s
    assert test3a.input() == "foo bar baz"

    # [4]
    # icu::RegexMatcher::RegexMatcher(
    #       UText *regexp,
    #       UText *input,
    #       uint32_t flags,
    #       UErrorCode &status
    # )
    regexp4 = utext_open_utf8(None, s, -1)
    src4 = utext_open_utf8(None, "foo bar baz", -1)
    test4 = RegexMatcher(regexp4, src4, URegexpFlag.UREGEX_CASE_INSENSITIVE)
    pattern4 = test4.pattern()
    assert pattern4.pattern() == s
    assert test4.input() == "foo bar baz"
    utext_close(regexp4)
    utext_close(src4)


def test_region():
    regexp = UnicodeString("\\w+")
    src = UnicodeString("foo bar baz")
    matcher = RegexMatcher(regexp, src, 0)
    assert matcher.region_start() == 0
    assert matcher.region_start64() == 0
    assert matcher.region_end() == len(src)
    assert matcher.region_end64() == len(src)
    assert matcher.find()
    assert matcher.group() == "foo"

    # [1]
    # RegexMatcher &icu::RegexMatcher::region(
    #       int64_t regionStart,
    #       int64_t regionLimit,
    #       int64_t startIndex,
    #       UErrorCode &status
    # )
    result = matcher.region(3, 7, 5)
    assert isinstance(result, RegexMatcher)
    assert result == matcher
    assert matcher.region_start() == 3
    assert matcher.region_start64() == 3
    assert matcher.region_end() == 7
    assert matcher.region_end64() == 7
    assert matcher.find()
    assert matcher.group() == "ar"

    # [2]
    # RegexMatcher &icu::RegexMatcher::region(
    #       int64_t start,
    #       int64_t limit,
    #       UErrorCode &status
    # )
    result = matcher.region(3, 7)
    assert isinstance(result, RegexMatcher)
    assert result == matcher
    assert matcher.region_start() == 3
    assert matcher.region_start64() == 3
    assert matcher.region_end() == 7
    assert matcher.region_end64() == 7
    assert matcher.find()
    assert matcher.group() == "bar"


def test_replace_all():
    regexp = UnicodeString("abc")
    src = UnicodeString(".abc..abc..")
    matcher = RegexMatcher(regexp, src, 0)

    # [1]
    # UnicodeString icu::RegexMatcher::replaceAll(
    #       const UnicodeString &replacement,
    #       UErrorCode &status
    # )
    result = matcher.replace_all(UnicodeString("xyz"))
    assert isinstance(result, UnicodeString)
    assert result == ".xyz..xyz.."

    result = matcher.replace_all("xyz")
    assert isinstance(result, UnicodeString)
    assert result == ".xyz..xyz.."

    # [2]
    # UText *icu::RegexMatcher::replaceAll(
    #       UText *replacement,
    #       UText *dest,
    #       UErrorCode &status
    # )
    replacement = utext_open_utf8(None, "xyz", -1)
    dest = matcher.replace_all(replacement, None)
    assert utext_extract(dest, 0, utext_native_length(dest)) == ".xyz..xyz.."

    utext_open_utf8(replacement, "ABC", -1)
    utext_replace(dest, 0, utext_native_length(dest), "", -1)
    result = matcher.replace_all(replacement, dest)
    assert result == dest
    assert utext_extract(dest, 0, utext_native_length(dest)) == ".ABC..ABC.."
    utext_close(replacement)
    utext_close(dest)


def test_replace_first():
    regexp = UnicodeString("abc")
    src = UnicodeString(".abc..abc..")
    matcher = RegexMatcher(regexp, src, 0)

    # [1]
    # UnicodeString icu::RegexMatcher::replaceFirst(
    #       const UnicodeString &replacement,
    #       UErrorCode &status
    # )
    result = matcher.replace_first(UnicodeString("xyz"))
    assert isinstance(result, UnicodeString)
    assert result == ".xyz..abc.."

    result = matcher.replace_first("xyz")
    assert isinstance(result, UnicodeString)
    assert result == ".xyz..abc.."

    # [2]
    # UText *icu::RegexMatcher::replaceFirst(
    #       UText *replacement,
    #       UText *dest,
    #       UErrorCode &status
    # )
    replacement = utext_open_utf8(None, "xyz", -1)
    dest = matcher.replace_first(replacement, None)
    assert utext_extract(dest, 0, utext_native_length(dest)) == ".xyz..abc.."

    utext_open_utf8(replacement, "ABC", -1)
    utext_replace(dest, 0, utext_native_length(dest), "", -1)
    result = matcher.replace_first(replacement, dest)
    assert result == dest
    assert utext_extract(dest, 0, utext_native_length(dest)) == ".ABC..abc.."
    utext_close(replacement)
    utext_close(dest)


def test_reset():
    regexp = UnicodeString("\\w+")
    src1 = UnicodeString("foo")
    matcher = RegexMatcher(regexp, src1, 0)
    assert matcher.find()
    assert matcher.group() == "foo"
    assert not matcher.find()

    # [1]
    # RegexMatcher &icu::RegexMatcher::reset()
    result = matcher.reset()
    assert isinstance(result, RegexMatcher)
    assert id(result) == id(matcher)
    assert matcher.find()
    assert matcher.group() == "foo"
    assert not matcher.find()

    # [2]
    # RegexMatcher &icu::RegexMatcher::reset(const UnicodeString &input)
    src2 = UnicodeString("foo bar")
    result = matcher.reset(src2)
    assert isinstance(result, RegexMatcher)
    assert id(result) == id(matcher)
    assert matcher.find()
    assert matcher.group() == "foo"
    assert matcher.find()
    assert matcher.group() == "bar"
    assert not matcher.find()

    # [3]
    # RegexMatcher &icu::RegexMatcher::reset(
    #       int64_t index,
    #       UErrorCode &status
    # )
    result = matcher.reset(4)
    assert isinstance(result, RegexMatcher)
    assert id(result) == id(matcher)
    assert matcher.find()
    assert matcher.group() == "bar"
    assert not matcher.find()

    # [4]
    # RegexMatcher &icu::RegexMatcher::reset(UText *input)
    src4 = utext_open_utf8(None, "abc xyz", -1)
    result = matcher.reset(src4)
    assert isinstance(result, RegexMatcher)
    assert id(result) == id(matcher)
    assert matcher.find()
    assert matcher.group() == "abc"
    assert matcher.find()
    assert matcher.group() == "xyz"
    assert not matcher.find()
    utext_close(src4)


def test_set_find_progress_callback():
    result1 = []

    # UBool URegexFindProgressCallback(
    #       const void *context,
    #       int64_t matchIndex
    # )
    def _find_progress_callback1(_context: object, _match_index: int) -> bool:
        assert _context is None
        nonlocal result1
        result1.append(_match_index)
        return True

    def _find_progress_callback2(_context: object, _match_index: int) -> bool:
        assert isinstance(_context, list)
        _context.append(_match_index)
        if _match_index >= 5:
            return False
        return True

    regexp = UnicodeString("abc")
    src = UnicodeString(".abc..abc..")
    matcher = RegexMatcher(regexp, src, 0)

    # void icu::RegexMatcher::getFindProgressCallback(
    #       URegexFindProgressCallback *&callback,
    #       const void *&context,
    #       UErrorCode &status
    # )
    callback0, context0 = matcher.get_find_progress_callback()
    assert isinstance(callback0, URegexFindProgressCallbackPtr)
    assert isinstance(context0, ConstVoidPtr)

    # void icu::RegexMatcher::setFindProgressCallback(
    #       URegexFindProgressCallback *callback,
    #       const void *context,
    #       UErrorCode &status
    # )
    callback1 = URegexFindProgressCallbackPtr(_find_progress_callback1)
    context1 = ConstVoidPtr(None)
    matcher.set_find_progress_callback(callback1, context1)
    assert matcher.find()
    assert matcher.find()
    assert not matcher.find()
    assert result1 == [1, 5, 6]

    result2 = []
    callback2 = URegexFindProgressCallbackPtr(_find_progress_callback2)
    context2 = ConstVoidPtr(result2)
    matcher.set_find_progress_callback(callback2, context2)
    assert matcher.find(0)
    with pytest.raises(ICUError) as exc_info:
        matcher.find()
    assert exc_info.value.args[0] == UErrorCode.U_REGEX_STOPPED_BY_CALLER
    assert result2 == [1, 5]

    result1.clear()
    result2.clear()
    matcher.set_find_progress_callback(callback0, context0)
    assert matcher.find(0)
    assert matcher.find()
    assert not matcher.find()
    assert len(result1) == len(result2) == 0


def test_set_match_callback():
    result1 = []

    # UBool URegexMatchCallback(const void *context, int32_t steps)
    def _match_callback1(_context: object, _steps: int) -> bool:
        assert _context is None
        nonlocal result1
        result1.append(_steps)
        return True

    def _match_callback2(_context: object, _steps: int) -> bool:
        assert isinstance(_context, list)
        _context.append(_steps)
        if _steps >= 5:
            return False
        return True

    regexp = UnicodeString("((.)+\\2)+x")
    matcher = RegexMatcher(regexp, 0)
    src = UnicodeString("aaaaaaaaaaaaaaaaaaaaaaab")

    # void icu::RegexMatcher::getMatchCallback(
    #       URegexMatchCallback *&callback,
    #       const void *&context,
    #       UErrorCode &status
    # )
    callback0, context0 = matcher.get_match_callback()
    assert isinstance(callback0, URegexMatchCallbackPtr)
    assert isinstance(context0, ConstVoidPtr)

    # void icu::RegexMatcher::setMatchCallback(
    #       URegexMatchCallback *callback,
    #       const void *context,
    #       UErrorCode &status
    # )
    callback1 = URegexMatchCallbackPtr(_match_callback1)
    context1 = ConstVoidPtr(None)
    matcher.set_match_callback(callback1, context1)
    matcher.reset(src)
    assert not matcher.matches()
    assert result1 == [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16]

    result2 = []
    callback2 = URegexMatchCallbackPtr(_match_callback2)
    context2 = ConstVoidPtr(result2)
    matcher.set_match_callback(callback2, context2)
    matcher.reset(src)
    with pytest.raises(ICUError) as exc_info:
        matcher.matches()
    assert exc_info.value.args[0] == UErrorCode.U_REGEX_STOPPED_BY_CALLER
    assert result2 == [1, 2, 3, 4, 5]

    result1.clear()
    result2.clear()
    matcher.set_match_callback(callback0, context0)
    matcher.reset(src)
    assert not matcher.matches()
    assert len(result1) == len(result2) == 0


def test_split():
    regexp = UnicodeString("\\s+")
    matcher = RegexMatcher(regexp, 0)
    src1 = UnicodeString("foo bar baz")

    # [1]
    # int32_t icu::RegexMatcher::split(
    #       const UnicodeString &input,
    #       UnicodeString dest[],
    #       int32_t destCapacity,
    #       UErrorCode &status
    # )
    dest1 = UnicodeStringVector(10)
    result = matcher.split(src1, dest1)
    assert result == 3
    assert dest1[0] == "foo"
    assert dest1[1] == "bar"
    assert dest1[2] == "baz"

    result = matcher.split(src1, dest1, 2)
    assert result == 2
    assert dest1[0] == "foo"
    assert dest1[1] == "bar baz"

    # [2]
    # int32_t icu::RegexMatcher::split(
    #       UText *input,
    #       UText *dest[],
    #       int32_t destCapacity,
    #       UErrorCode &status
    # )
    src2 = utext_open_const_unicode_string(None, src1)
    dest2 = UTextVector(10)
    result = matcher.split(src2, dest2)
    assert result == 3
    assert utext_extract(dest2[0], 0, utext_native_length(dest2[0])) == "foo"
    assert utext_extract(dest2[1], 0, utext_native_length(dest2[1])) == "bar"
    assert utext_extract(dest2[2], 0, utext_native_length(dest2[2])) == "baz"

    result = matcher.split(src2, dest2, 2)
    assert result == 2
    assert utext_extract(dest2[0], 0, utext_native_length(dest2[0])) == "foo"
    assert (utext_extract(dest2[1], 0, utext_native_length(dest2[1]))
            == "bar baz")
    utext_close(src2)


def test_start():
    regexp = UnicodeString("01(23(45)67)(.*)")
    src = UnicodeString("0123456789")
    matcher = RegexMatcher(regexp, src, 0)

    matcher.looking_at(0)

    # [1]
    # int32_t icu::RegexMatcher::start(
    #       int32_t group,
    #       UErrorCode &status
    # )
    assert matcher.start(0) == 0
    assert matcher.start(1) == 2
    assert matcher.start(2) == 4
    assert matcher.start(3) == 8

    # int64_t icu::RegexMatcher::start64(
    #       int32_t group,
    #       UErrorCode &status
    # )
    assert matcher.start64(0) == 0
    assert matcher.start64(1) == 2
    assert matcher.start64(2) == 4
    assert matcher.start64(3) == 8

    # [2]
    # int32_t icu::RegexMatcher::start(UErrorCode &status)
    assert matcher.start() == 0

    # int64_t icu::RegexMatcher::start64(UErrorCode &status)
    assert matcher.start64() == 0
